/*
 * Developed by Hans-Henning Ramberger as part of a master thesis.
 * Copyright (c) 2018.
 */

package de.uniko.bks.cli;

import de.uniko.bks.generator.entry.OutputEntryList;
import de.uniko.bks.statistic.PerformanceTestConfiguration;
import java.util.Arrays;

/** Wrapper for all available cli arguments of the DMN gemerator */
public class CliArgument {
  private static final String FLAG_EXCEPTION_MESSAGE =
      "Ignoring invalid or incomplete parameter and using the default value.";
  /** Sign that all flags have */
  private static String FLAG_SIGN = "-";
  /** Flag which identifies the row number steps */
  private static final String ROW_NUMBER_STEPS_FLAG = FLAG_SIGN + "r";

  private static final String NUMBER_OF_TEST_RUNS_FLAG = FLAG_SIGN + "t";
  private static final String NUMBER_OF_DATA_POINTS_FLAG = FLAG_SIGN + "d";
  private static final String INITIAL_COLUMN_COUNT_FLAG = FLAG_SIGN + "icc";
  private static final String INITIAL_ROW_COUNT_FLAG = FLAG_SIGN + "irc";
  private static final String OUTPUT_ENTRY_LIST_FLAG = FLAG_SIGN + "o";
  /** Difference in rows in each test dmn file */
  private int rowSteps;
  /** Number of runs for each file to calculate the average run-time */
  private int numberOfTestRuns;
  /** Amount (square root) of different data points (files) created */
  private int numberOfDataPoints;
  /** Number of columns the generator starts with */
  private int initialColumnCount;
  /** Number of rows the generator starts with */
  private int initialRowCount;
  /** List of output entries used in the dmn file (note the negations are created automatically) */
  private OutputEntryList outputEntryList = new OutputEntryList();

  public CliArgument(String[] args) {
    this.rowSteps =
        this.getFlagValue(
            ROW_NUMBER_STEPS_FLAG, args, PerformanceTestConfiguration.getRowNumberSteps());
    this.numberOfTestRuns =
        this.getFlagValue(
            NUMBER_OF_TEST_RUNS_FLAG, args, PerformanceTestConfiguration.getNumberOfTests());
    this.numberOfDataPoints =
        this.getFlagValue(
            NUMBER_OF_DATA_POINTS_FLAG, args, PerformanceTestConfiguration.getNumberOfDataPoints());
    this.initialColumnCount =
        this.getFlagValue(
                INITIAL_COLUMN_COUNT_FLAG,
                args,
                PerformanceTestConfiguration.getInitialColumnCount())
            - 1;
    this.initialRowCount =
        this.getFlagValue(
                INITIAL_ROW_COUNT_FLAG, args, PerformanceTestConfiguration.getInitialRowCount())
            - 1;
    this.outputEntryList.setValues(
        this.getFlagValue(OUTPUT_ENTRY_LIST_FLAG, args, new OutputEntryList().getValues()));
  }

  /**
   * Function which gets the value of a cli flag
   *
   * @param flag Flag of which the values is extracted
   * @param args All cli arguments given
   * @param defaultValue Default value if the argument is not set correctly
   * @return The correct value of the cli argument
   */
  private int getFlagValue(String flag, String[] args, int defaultValue) {
    try {
      int flagIndex = Arrays.asList(args).indexOf(flag);
      // Check if flag is set and the next item is not a flag
      if (flagIndex >= 0 && !args[flagIndex + 1].contains(FLAG_SIGN)) {
        return Integer.parseInt(args[flagIndex + 1]);
      } else {
        // Set the default value for the row steps
        return defaultValue;
      }
    } catch (ArrayIndexOutOfBoundsException | NumberFormatException e) {
      System.err.println(FLAG_EXCEPTION_MESSAGE);
    }
    // Set the default value for the parameter
    return defaultValue;
  }

  /**
   * Function which gets the value of a cli flag
   *
   * @param flag Flag of which the values is extracted
   * @param args All cli arguments given
   * @param defaultValue Default value if the argument is not set correctly
   * @return The correct value of the cli argument
   */
  private String[] getFlagValue(String flag, String[] args, String[] defaultValue) {
    try {
      int flagIndex = Arrays.asList(args).indexOf(flag);
      // Check if flag is set and the next item is not a flag
      if (flagIndex >= 0 && !args[flagIndex + 1].contains(FLAG_SIGN)) {
        // Split the comma separated items of the flag
        return args[flagIndex + 1].split(",");
      } else {
        // Set the default value for the row steps
        return defaultValue;
      }
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println(FLAG_EXCEPTION_MESSAGE);
    }
    // Set the default value for the parameter
    return defaultValue;
  }

  public int getRowSteps() {
    return rowSteps;
  }

  public int getNumberOfTestRuns() {
    return numberOfTestRuns;
  }

  public int getNumberOfDataPoints() {
    return numberOfDataPoints;
  }

  public int getInitialColumnCount() {
    return initialColumnCount;
  }

  public int getInitialRowCount() {
    return initialRowCount;
  }

  public OutputEntryList getOutputEntryList() {
    return outputEntryList;
  }
}
