/*
 * Developed by Hans-Henning Ramberger as part of a master thesis.
 * Copyright (c) 2018.
 */

package de.uniko.bks.dom.dfl.rule;

import de.uniko.bks.util.StringSanitizer;
import java.util.Comparator;

/** Class representing a DFL rule identifier */
public class RuleId {
  /** Delimiter between two parts of the ruleId */
  private static final String PART_DELIMITER =
      RuleDelimiter.WHITESPACE.getValue()
          + RuleDelimiter.WHITESPACE.getValue()
          + RuleDelimiter.WHITESPACE.getValue();

  /** XML id of the rule in the DMN */
  private String xmlId;
  /** Id of the decision in the DMN */
  private String decisionId;
  /** Number of the rule */
  private int number;
  /** Subnumber of the rule (only important on split up rules) */
  private int subNumber;
  /** Priority of the rule */
  private int priority;

  /**
   * Constructor to completely initialize a RuleId
   *
   * @param xmlId XML id of the rule in the DMN
   * @param decisionId Id of the decision in the DMN
   * @param number Number of the rule
   * @param subNumber Subnumber of the rule (only important on split up rules)
   * @param priority Priority of the rule
   */
  public RuleId(String xmlId, String decisionId, int number, int subNumber, int priority) {
    // Replace illegal characters
    this.xmlId = StringSanitizer.sanitize(xmlId, true, false, false, true);
    this.decisionId = decisionId;
    this.number = number;
    this.subNumber = subNumber;
    this.priority = priority;
  }

  /**
   * Getter for the XML id of the rule in the DMN
   *
   * @return The XML id of the rule in the DMN
   */
  public String getXmlId() {
    return xmlId;
  }

  /**
   * Getter for the id of the decision in the DMN
   *
   * @return The id of the decision in the DMN
   */
  public String getDecisionId() {
    return decisionId;
  }

  /**
   * Getter for the number of the rule
   *
   * @return The number of the rule
   */
  public int getNumber() {
    return number;
  }

  /**
   * Getter for the subnumber of the rule (only important on split up rules)
   *
   * @return The subnumber of the rule
   */
  public int getSubNumber() {
    return subNumber;
  }

  /**
   * Getter for the priority of the rule
   *
   * @return The priority of the rule
   */
  public int getPriority() {
    return priority;
  }

  /**
   * Getter for the delimiter between two parts of the ruleId
   *
   * @return The delimiter between two parts of the ruleId
   */
  public static String getPartDelimiter() {
    return PART_DELIMITER;
  }

  /** Define the comparator to sort the ruleIds by their number */
  public static Comparator<RuleId> COMPARE_BY_NUMBER = Comparator.comparingInt(one -> one.number);

  /** Define the comparator to sort th ruleIds by their priority */
  public static Comparator<RuleId> COMPARE_BY_PRIORITY =
      Comparator.comparingInt(one -> one.priority);

  @Override
  public String toString() {
    // Do not render the xml id on facts
    if (xmlId.equals("")) {
      return this.decisionId + PART_DELIMITER + this.number + RuleDelimiter.ID.getValue();
    }

    return this.xmlId
        + PART_DELIMITER
        + this.decisionId
        + PART_DELIMITER
        + this.number
        + PART_DELIMITER
        + this.subNumber
        + PART_DELIMITER
        + this.priority
        + RuleDelimiter.ID.getValue();
  }
}
