/*
 * Developed by Hans-Henning Ramberger as part of a master thesis.
 * Copyright (c) 2018.
 */

package de.uniko.bks.dom.dmn.entry;

/**
 * Class specifying a common transfer object encapsulating a rule entry
 *
 * @param <T> Type (java class) of the RuleEntry value
 */
public class RuleEntry<T> {
  /** Type (java class) of the value */
  private String type;
  /** Value of the RuleEntry */
  private T value;
  /** Name (i.e. DMN table header) of the RuleEntry */
  private String name;
  /** Whether the values is negated in a logical sense or not */
  private boolean negated;
  /**
   * Whether the RuleEntry is an output entry (this property is needed to avoid issues with
   * SPINdle's way of handling head literals)
   */
  private boolean output;

  /**
   * Constructor for a simple non-negated RuleEntry
   *
   * @param type Type (java class) of the value
   * @param value Value of the RuleEntry
   * @param headerName Name (i.e. DMN table header) of the RuleEntry
   */
  public RuleEntry(String type, T value, String headerName) {
    this.type = type;
    this.value = value;
    this.name = headerName;
    this.negated = false;
    this.output = false;
  }

  /**
   * Constructor for a possible negated RuleEntry
   *
   * @param type Type (java class) of the value
   * @param value Value of the RuleEntry
   * @param headerName Name (i.e. DMN table header) of the RuleEntry
   * @param negated Whether the values is negated in a logical sense or not
   */
  public RuleEntry(String type, T value, String headerName, boolean negated) {
    this.type = type;
    this.value = value;
    this.name = headerName;
    this.negated = negated;
    this.output = false;
  }

  /**
   * Constructor for an RuleEntry which is an dmn output
   *
   * @param type Type (java class) of the value
   * @param value Value of the RuleEntry
   * @param headerName Name (i.e. DMN table header) of the RuleEntry
   * @param negated Whether the values is negated in a logical sense or not
   * @param output Whether the RuleEntry is an output entry
   */
  public RuleEntry(String type, T value, String headerName, boolean negated, boolean output) {
    this.type = type;
    this.value = value;
    this.name = headerName;
    this.negated = negated;
    this.output = output;
  }

  /**
   * Getter for the RuleEntry type
   *
   * @return The type of the RuleEntry
   */
  public String getType() {
    return type;
  }

  /**
   * Getter for the RuleEntry value
   *
   * @return The value of the RuleEntry
   */
  public T getValue() {
    return value;
  }

  /**
   * Setter for the RuleEntry value
   *
   * @param value The value of the RuleEntry
   */
  public void setValue(T value) {
    this.value = value;
  }

  /**
   * Getter for the RuleEntry name
   *
   * @return The name of the RuleEntry
   */
  public String getName() {
    return name;
  }

  /**
   * Getter for the RuleEntry negated property
   *
   * @return Whether or not the RuleEntry is negated
   */
  public boolean isNegated() {
    return negated;
  }

  /**
   * Getter for the RuleEntry output property
   *
   * @return Whether or not the RuleEntry is an output
   */
  public boolean isOutput() {
    return output;
  }

  /**
   * Setter for the RuleEntry output property
   *
   * @param output Whether or not the RuleEntry is an output
   */
  public void setOutput(boolean output) {
    this.output = output;
  }
}
