/*
 * Developed by Hans-Henning Ramberger as part of a master thesis.
 * Copyright (c) 2018.
 */

package de.uniko.bks.generator;

import de.uniko.bks.dom.dmn.Decision;
import de.uniko.bks.dom.dmn.DecisionTable;
import de.uniko.bks.dom.dmn.DmnRule;
import de.uniko.bks.dom.dmn.entry.InputEntry;
import de.uniko.bks.generator.entry.InputEntryGenerator;
import de.uniko.bks.generator.entry.OutputEntryGenerator;
import de.uniko.bks.generator.entry.OutputEntryList;
import de.uniko.bks.generator.header.InputHeaderGenerator;
import de.uniko.bks.generator.header.OutputHeaderGenerator;
import de.uniko.bks.xml.FileHandler;
import de.uniko.bks.xml.XmlMarshaller;
import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.UUID;
import javax.xml.bind.JAXBException;

/** Main class of the generator which supervises the DMN generation */
public class DmnGenerator {
  /** Identifier suffix for all ids in the DMN decision */
  private static String identifier = UUID.randomUUID().toString().substring(0, 4);

  /**
   * Function which generates a DMN rule list
   *
   * @param columnCount Number of columns in the DMN table
   * @param rowCount Number of rows in the DMN table
   * @param outputEntryList List of output entries used in the dmn file (note the negations are
   *     created automatically)
   * @return The DMN rule list
   */
  private static ArrayList<DmnRule> generateDmnRuleList(
      int columnCount, int rowCount, OutputEntryList outputEntryList) {
    ArrayList<DmnRule> dmnRuleList = new ArrayList<>();
    // Generate a rule for each row
    for (int i = 0; i < rowCount; i++) {
      DmnRule dmnRule = new DmnRule();
      dmnRule.setId("dmnRule_" + identifier + "_" + i);
      // Generate literals for a rule
      ArrayList<InputEntry> inputEntryList = new ArrayList<>();
      for (int j = 0; j < columnCount; j++) {
        inputEntryList.add(InputEntryGenerator.generateInputEntry(identifier, j, i));
      }
      dmnRule.setInputEntry(inputEntryList);
      dmnRule.setOutputEntry(
          OutputEntryGenerator.generateOutputEntry(identifier, i, outputEntryList));
      dmnRuleList.add(dmnRule);
    }

    return dmnRuleList;
  }

  /**
   * Function which generates a DMN decsion table
   *
   * @param columnCount Number of columns in the DMN table
   * @param rowCount Number of rows in the DMN table
   * @param hitPolicy Hit policy for the DMN table
   * @param outputEntryList List of output entries used in the dmn file (note the negations are
   *     created automatically)
   * @return The generated DMN decision table
   */
  private static DecisionTable generateDecisionTable(
      int columnCount, int rowCount, String hitPolicy, OutputEntryList outputEntryList) {
    DecisionTable decisionTable = new DecisionTable();
    decisionTable.setId("decisionTable_" + identifier);
    decisionTable.setHitPolicy(hitPolicy);
    decisionTable.setInput(InputHeaderGenerator.generateInputList(identifier, columnCount));
    decisionTable.setOutputHeader(
        OutputHeaderGenerator.generateOutputHeader(identifier, hitPolicy, outputEntryList));
    decisionTable.setDmnRule(generateDmnRuleList(columnCount, rowCount, outputEntryList));
    return decisionTable;
  }

  /**
   * Function which generates a DMN file
   *
   * @param columnCount Number of columns in the DMN table
   * @param rowCount Number of rows in the DMN table
   * @param hitPolicy Hit policy for the DMN table
   * @param outputEntryList List of output entries used in the dmn file (note the negations are
   *     created automatically)
   * @return The path of the newly created file
   * @throws JAXBException Root exception for all JAXB exceptions
   * @throws IOException Signals that an I/O exception of some sort has occurred
   */
  public static Path generateDmnFile(
      int columnCount, int rowCount, String hitPolicy, OutputEntryList outputEntryList)
      throws JAXBException, IOException {
    Decision decision = new Decision();
    decision.setId("decision_" + identifier);
    decision.setName("decision_" + identifier);
    decision.setDmnTable(generateDecisionTable(columnCount, rowCount, hitPolicy, outputEntryList));

    // Generate the XML fragment
    System.out.println(
        "Generating DMN file with "
            + columnCount
            + " columns, "
            + rowCount
            + " rows and the "
            + hitPolicy
            + " hit policy");
    String xmlResult = XmlMarshaller.marshalDmnTable(decision);
    // Get the template for an empty DMN model
    String dmnTemplate = FileHandler.read("EmptyDmn.xml");
    // Replace the placeholder with the generated content
    String fileContent = dmnTemplate.replace("<!--replace-me-->", xmlResult);
    // Write the content into a new file
    Path filePath = FileHandler.write(identifier, columnCount, rowCount, hitPolicy, fileContent);
    System.out.println("Created new DMN file successfully with the path " + filePath.toString());
    return filePath;
  }
}
