/*
 * Developed by Hans-Henning Ramberger as part of a master thesis.
 * Copyright (c) 2018.
 */

package de.uniko.bks.statistic;

import de.uniko.bks.cli.CliArgument;
import de.uniko.bks.generator.DmnGenerator;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import javax.xml.bind.JAXBException;

/** Class which processes automatic performance tests of the converter */
public class PerformanceTester {

  private static CliArgument cliArgument;

  /**
   * Function which calculates the average run-time of the conversion for one file
   *
   * @param runTimeList List of run-times for one test run
   * @return The average run-time for the conversion of one file in milliseconds
   */
  private static double calculateAverageRunTime(ArrayList<Long> runTimeList) {
    double totalRunTime = 0;
    // Add upt the run-time from each run
    for (Long runTime : runTimeList) {
      totalRunTime += runTime;
    }
    // Calculate the average of the runtime
    return (totalRunTime / cliArgument.getNumberOfTestRuns());
  }

  /**
   * Function which runs the performance tests
   *
   * @param dmnFileList List of DMN files to process
   * @return The average run-time of a conversion of one file
   * @throws IOException Signals that an I/O exception of some sort has occurred
   */
  private static ArrayList<ArrayList<ParserStatistic>> runPerformanceTests(
      ArrayList<DmnFile> dmnFileList) throws IOException {
    ArrayList<ArrayList<ParserStatistic>> parserStatisticList = new ArrayList<>();

    ArrayList<ParserStatistic> statistics = new ArrayList<>();
    ArrayList<ParserStatistic> verifierStatistics = new ArrayList<>();
    int size = dmnFileList.size() * cliArgument.getNumberOfTestRuns();
    int j = 1;
    for (DmnFile dmnFile : dmnFileList) {
      ArrayList<Long> runTimeList = new ArrayList<>();
      ArrayList<Long> verificationsFound = new ArrayList<>();
      for (int i = 0; i < cliArgument.getNumberOfTestRuns(); i++) {
        TestDMN testDMN = new TestDMN();
        // Add the run-time of the current execution to the list of run-times for the file
        testDMN.testDmn(dmnFile.getPath().toString());
        verificationsFound.add(
            testDMN
                .getVerifcation()
                .getVerifierResults()
                .stream()
                .map(verifierResult -> verifierResult.getVerifications().size())
                .mapToLong(Long::valueOf)
                .sum());
        runTimeList.add(testDMN.getExecTime());
        System.out.println(
            "DMN " + j + " von " + size + " verified in " + testDMN.getExecTime() + "ms");
        j++;
      }
      statistics.add(
          new ParserStatistic(
              dmnFile.getHitPolicy(),
              dmnFile.getColumnCount(),
              dmnFile.getRowCount(),
              // Calculate the average run-time for the current file
              calculateAverageRunTime(runTimeList)));
      verifierStatistics.add(
          new ParserStatistic(
              "Count verifications found",
              dmnFile.getColumnCount(),
              dmnFile.getRowCount(),
              // Calculate the average run-time for the current file
              calculateAverageRunTime(verificationsFound)));
    }
    parserStatisticList.add(statistics);
    parserStatisticList.add(verifierStatistics);
    return parserStatisticList;
  }

  private static String toTikz(ArrayList<ParserStatistic> parserStatisticList) {
    String newline = System.getProperty("line.separator");
    StringBuilder tikzCode = new StringBuilder();
    tikzCode.append("\\begin{figure}[h!]").append(newline);
    tikzCode.append(" \\centering").append(newline);
    tikzCode.append(" \\begin{tikzpicture}").append(newline);
    tikzCode
        .append("     \\begin{axis}")
        .append(
            "[xlabel=Number of columns, ylabel=Number of rows, zlabel=Run-time in milliseconds, xlabel style={sloped like x axis}, ylabel style={sloped}, view/az=45, zmin=0, grid=major]")
        .append(newline);
    tikzCode.append("         \\addplot3[surf, shader=faceted]").append(newline);
    tikzCode.append("         coordinates {").append(newline);

    // Add the coordinates (x=columnCount, y=rowCount, z=averageRunTime)
    int i = 1;
    for (ParserStatistic parserStatistic : parserStatisticList) {
      // Create a matrix NUMBER_OF_DATA_POINTS X NUMBER_OF_DATA_POINTS
      if (i == 1) {
        tikzCode.append("         ");
      }
      tikzCode.append("(");
      tikzCode.append(parserStatistic.getColumnCount()).append(", ");
      tikzCode.append(parserStatistic.getRowCount()).append(", ");
      tikzCode.append(parserStatistic.getAverageRunTime()).append(") ");

      if (i % cliArgument.getNumberOfDataPoints() == 0 && i < parserStatisticList.size()) {
        tikzCode.append(newline).append(newline).append("         ");
      }
      i++;
    }

    tikzCode.append(newline).append("       };").append(newline);
    tikzCode.append("  \\end{axis}").append(newline);
    tikzCode.append(" \\end{tikzpicture}").append(newline);
    tikzCode
        .append("\\caption{Run-time statistics for the ")
        .append(parserStatisticList.get(0).getHitPolicy())
        .append(" hit policy} ")
        .append(newline);
    tikzCode
        .append("\\label{fig:statistic-")
        .append(parserStatisticList.get(0).getHitPolicy().toLowerCase().replace(" ", "-"))
        .append("}")
        .append(newline);
    tikzCode.append("\\end{figure}").append(newline);

    return tikzCode.toString();
  }

  /**
   * Function which generates the DMN files analyzed for performance
   *
   * @param hitPolicy Hit policy for the DMN table
   * @return List of DMN files which should be tested
   * @throws JAXBException Root exception for all JAXB exceptions
   * @throws IOException Signals that an I/O exception of some sort has occurred
   */
  private static ArrayList<DmnFile> generateTestFiles(String hitPolicy, int rowNumberSteps)
      throws JAXBException, IOException {
    ArrayList<DmnFile> dmnFileList = new ArrayList<>();
    // Create the given amount of files
    for (int i = 1; i <= cliArgument.getNumberOfDataPoints(); i++) {
      // Create the file
      for (int j = 1; j <= cliArgument.getNumberOfDataPoints(); j++) {
        Path filePath =
            (DmnGenerator.generateDmnFile(
                i + cliArgument.getInitialColumnCount(),
                (rowNumberSteps * j) + cliArgument.getInitialRowCount(),
                hitPolicy,
                cliArgument.getOutputEntryList()));

        // Add the file to the list of generated files
        dmnFileList.add(
            new DmnFile(
                hitPolicy,
                filePath,
                i + cliArgument.getInitialColumnCount(),
                (cliArgument.getRowSteps() * j) + cliArgument.getInitialRowCount()));
      }
    }
    return dmnFileList;
  }

  /**
   * Function which deletes a list of DMN files
   *
   * @param dmnFileList List of DMN files which should be deleted
   * @throws IOException Signals that an I/O exception of some sort has occurred
   */
  private static void deleteTestFiles(ArrayList<DmnFile> dmnFileList) throws IOException {
    for (DmnFile dmnFile : dmnFileList) {
      Files.deleteIfExists(dmnFile.getPath());
    }
    // Add an information to the cli
    System.out.println("Deleted " + dmnFileList.size() + " files successfully");
  }

  public static void main(String argv[]) throws JAXBException, IOException {
    cliArgument = new CliArgument(argv);
    // Measure the start time
    RunTimePrinter runTimePrinter = new RunTimePrinter(System.currentTimeMillis());

    ArrayList<ArrayList<DmnFile>> dmnFileLists = new ArrayList<>();
    ArrayList<String> tikzPictureList = new ArrayList<>();
    // Generate the DMN files
    dmnFileLists.add(generateTestFiles("COLLECT", cliArgument.getRowSteps()));
    // Run the performance tests
    for (ArrayList<DmnFile> dmnFileList : dmnFileLists) {
      ArrayList<ArrayList<ParserStatistic>> parserStatisticList = runPerformanceTests(dmnFileList);
      // Generate tikz code
      for (ArrayList<ParserStatistic> set : parserStatisticList) {
        tikzPictureList.add(toTikz(set));
      }

      // Delete the generated test files if needed
      if (PerformanceTestConfiguration.isDeleteTestFiles()) {
        deleteTestFiles(dmnFileList);
      }
    }

    // Print the tikz code of the statistics
    tikzPictureList.forEach(System.out::println);

    // Measure the end time
    runTimePrinter.setEndTime(System.currentTimeMillis());
    // Print the total execution time
    System.out.println(runTimePrinter.toString());
  }
}
