/*
 * Developed by Hans-Henning Ramberger as part of a master thesis.
 * Copyright (c) 2018.
 */

package de.uniko.bks.util;

import de.uniko.bks.dom.dfl.rule.Comparator;
import de.uniko.bks.dom.dfl.rule.RuleDelimiter;

/** Class providing functions to sanitize strings */
public class StringSanitizer {
  /**
   * Function which removes whitespaces from a given text
   *
   * @param text Text in which all whitespaces should be removed
   * @return The sanitized text
   */
  private static String replaceWhitespace(String text) {
    text = text.replace(Comparator.SMALLER.getValue() + " ", Comparator.SMALLER.getValue());
    text =
        text.replace(
            Comparator.SMALLER_OR_EQUALS.getValue() + " ", Comparator.SMALLER_OR_EQUALS.getValue());
    text =
        text.replace(
            Comparator.GREATER_OR_EQUALS.getValue() + " ", Comparator.GREATER_OR_EQUALS.getValue());
    text = text.replace(Comparator.GREATER.getValue() + " ", Comparator.GREATER.getValue());

    return text.replace(" ", RuleDelimiter.WHITESPACE.getValue());
  }

  /**
   * Function which removes double quotes from a given text
   *
   * @param text Text in which all double quotes should be removed
   * @return The sanitized text
   */
  private static String replaceQuotes(String text) {
    return text.replace("\"", "");
  }

  /**
   * Function which removes colons from a given text
   *
   * @param text Text in which all colons should be removed
   * @return The sanitized text
   */
  private static String replaceColon(String text) {
    return text.replace(":", "");
  }

  /**
   * Function which removes dashes from a given text
   *
   * @param text Text in which all dashes should be removed
   * @return The sanitized text
   */
  private static String replaceDash(String text) {
    return text.replace(
        "-", RuleDelimiter.WHITESPACE.getValue() + RuleDelimiter.WHITESPACE.getValue());
  }

  /**
   * Function which sanitizes whitespaces
   *
   * @param text Text in which all whitespaces should be sanitized
   * @return The sanitized text
   */
  public static String sanitize(String text) {
    return sanitize(text, true, true, true, true);
  }

  /**
   * Function which sanitizes whitespaces
   *
   * @param text Text in which all whitespaces should be sanitized
   * @param replaceWhitespace Whether or not whitespaces should be sanitized
   * @return The sanitized text
   */
  public static String sanitize(String text, boolean replaceWhitespace) {
    return sanitize(text, replaceWhitespace, false, false, false);
  }

  /**
   * Function which sanitizes whitespaces and double quotes
   *
   * @param text Text which should be sanitized
   * @param replaceWhitespace Whether or not whitespaces should be sanitized
   * @param replaceQuotes Whether or not double quotes should be sanitized
   * @return The sanitized text
   */
  public static String sanitize(String text, boolean replaceWhitespace, boolean replaceQuotes) {
    return sanitize(text, replaceWhitespace, replaceQuotes, false, false);
  }

  /**
   * Function which sanitizes whitespaces, double quotes and colons
   *
   * @param text Text which should be sanitized
   * @param replaceWhitespace Whether or not whitespaces should be sanitized
   * @param replaceQuotes Whether or not double quotes should be sanitized
   * @param replaceColon Whether or not colons should be sanitized
   * @return The sanitized text
   */
  public static String sanitize(
      String text, boolean replaceWhitespace, boolean replaceQuotes, boolean replaceColon) {
    return sanitize(text, replaceWhitespace, replaceQuotes, replaceColon, false);
  }

  /**
   * Function which sanitizes whitespaces, double quotes and colons
   *
   * @param text Text which should be sanitized
   * @param replaceWhitespace Whether or not whitespaces should be sanitized
   * @param replaceQuotes Whether or not double quotes should be sanitized
   * @param replaceColon Whether or not colons should be sanitized
   * @param replaceDash Whether or not dashes should be sanitized
   * @return The sanitized text
   */
  public static String sanitize(
      String text,
      boolean replaceWhitespace,
      boolean replaceQuotes,
      boolean replaceColon,
      boolean replaceDash) {
    if (replaceWhitespace) {
      text = replaceWhitespace(text);
    }
    if (replaceQuotes) {
      text = replaceQuotes(text);
    }
    if (replaceColon) {
      text = replaceColon(text);
    }
    if (replaceDash) {
      text = replaceDash(text);
    }
    return text;
  }

  /**
   * Function removes whitespaces after commas
   *
   * @param text Text in which all whitespaces after commas should be removed
   * @return The text without whitespaces after commas
   */
  public static String removeWhitespacesAfterComma(String text) {
    return text.replaceAll(",\\s+", ",");
  }
}
