/*
 * Developed by Hans-Henning Ramberger as part of a master thesis.
 * Copyright (c) 2018.
 */

package de.uniko.bks.xml;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.stream.Collectors;

/** Class which provides basic functionalities for handling files */
public class FileHandler {
  private static final String FILE_EXTENSION = ".xml";

  /**
   * Function which reads a file into a variable
   *
   * @param fileName Name of the file which should be read
   * @return The content of the given file
   */
  public static String read(String fileName) {
    // Get the class loader to load the resource
    ClassLoader classLoader = FileHandler.class.getClassLoader();
    InputStream inputStream = classLoader.getResourceAsStream(fileName);

    // Convert the stream into a String
    return new BufferedReader(new InputStreamReader(inputStream))
        .lines()
        .collect(Collectors.joining("\n"));
  }

  /**
   * Function which reads a file into a variable
   *
   * @param filePath Complete path of the file which should be read
   * @return The content of the given file
   * @throws IOException Signals that an I/O exception of some sort has occurred
   */
  public static String read(Path filePath) throws IOException {
    return new String(Files.readAllBytes(filePath), StandardCharsets.UTF_8);
  }

  /**
   * Function which generates a name for the DMN file
   *
   * @param identifier Identifier suffix for all ids in the DMN decision
   * @param columnCount Number of columns in the DMN table
   * @param rowCount Number of rows in the DMN table
   * @param hitPolicy Hit policy for the DMN table
   * @return The name of the generated DMN file
   */
  private static String generateFileName(
      String identifier, int columnCount, int rowCount, String hitPolicy) {
    return "generated_dmn_c"
        + columnCount
        + "_r"
        + rowCount
        + "_h"
        + hitPolicy.replace(" ", "_")
        + "_"
        + identifier;
  }

  /**
   * Function which writes a given content into a new file
   *
   * @param identifier Identifier suffix for all ids in the DMN decision
   * @param columnCount Number of columns in the DMN table
   * @param rowCount Number of rows in the DMN table
   * @param content The content which is written into the new file
   * @param hitPolicy Hit policy for the DMN table
   * @return The path of the new file
   * @throws IOException Signals that an I/O exception of some sort has occurred
   */
  public static Path write(
      String identifier, int columnCount, int rowCount, String hitPolicy, String content)
      throws IOException {
    // Write the file to the temporary directory of the os
    File file =
        File.createTempFile(
            generateFileName(identifier, columnCount, rowCount, hitPolicy), FILE_EXTENSION);

    // Fill the generated file with content
    return Files.write(file.toPath(), content.getBytes());
  }
}
