package de.unikoblenz.fgbks.base.wordnet;

import edu.mit.jwi.Dictionary;
import edu.mit.jwi.IDictionary;
import edu.mit.jwi.item.IIndexWord;
import edu.mit.jwi.item.ISynset;
import edu.mit.jwi.item.IWord;
import edu.mit.jwi.item.IWordID;
import edu.mit.jwi.item.POS;
import java.io.File;
import java.net.URL;
import java.util.List;

/**
 * WordnetService provides some functions from wordnet <br>
 * First, get the instance with getInstance() (singleton) and than use the provided functions.
 */
public class WordnetService {

  private static final String PATH_WORDNET = "dict/";
  private static WordnetService instance;

  private IDictionary dictionary;

  private WordnetService() {
    setupDictionary();
  }

  // Singleton pattern

  /**
   * Get the singleton instance of the wordnet service.
   *
   * @param refresh true for a new initialisation of wordnet
   * @return the wordnet service instance
   */
  public static WordnetService getInstance(boolean refresh) {
    if (instance == null) {
      instance = new WordnetService();
    } else if (refresh) {
      instance.setupDictionary();
    }
    return instance;
  }

  /**
   * Get the singleton instance of the wordnet service.
   *
   * @return the wordnet service instance
   */
  public static WordnetService getInstance() {
    return getInstance(false);
  }

  private void setupDictionary() {
    if (dictionary != null) {
      dictionary.close();
    }
    dictionary = null;
    try {
      String totalPath =
          new File(getClass().getClassLoader().getResource(PATH_WORDNET).getFile())
              .getAbsolutePath();
      dictionary = new Dictionary(new URL("file", null, totalPath));
      dictionary.open();
    } catch (Exception e) {
      dictionary = null;
    }
  }

  /**
   * Return true, if both nouns are in one synset. <br>
   * "examination" & "exam" = true <br>
   * "bill" & "invoice" = true<br>
   *
   * @return true, if both nouns are synonyms
   */
  public boolean areNounsSynonyms(String noun1, String noun2) {
    if (dictionary == null) {
      return false;
    }
    noun1 = noun1.toLowerCase();
    noun2 = noun2.toLowerCase();
    IIndexWord idxWord = null;
    try {
      idxWord = dictionary.getIndexWord(noun1, POS.NOUN);
    } catch (Exception e) {
    }
    if (idxWord == null) {
      return false;
    }
    List<IWordID> wordIDs = idxWord.getWordIDs();
    for (IWordID wID : wordIDs) {
      IWord word = dictionary.getWord(wID);
      ISynset synset = word.getSynset();

      // iterate over words associated with the synset
      for (IWord w : synset.getWords()) {
        if (w.getLemma().equals(noun2)) {
          return true;
        }
      }
    }
    return false;
  }

  public void close() {
    if (dictionary != null && dictionary.isOpen()) {
      dictionary.close();
    }
  }
}
