package de.unikoblenz.fgbks.dmn.core.dmnhelper;

import static de.unikoblenz.fgbks.dmn.core.dmnhelper.Boundary.BoundType.EXCLUSIVE;
import static de.unikoblenz.fgbks.dmn.core.dmnhelper.Boundary.BoundType.INCLUSIVE;
import static java.lang.Double.NEGATIVE_INFINITY;
import static java.lang.Double.POSITIVE_INFINITY;

import de.unikoblenz.fgbks.base.builder.DefaultBuilder;
import java.util.Objects;
import java.util.Optional;

/** Class for manage boundaries and intervals. Can also handle String values. */
public class Boundary implements Comparable<Boundary>, Cloneable {

  public static final Boundary INFINITY_BOUNDARY = Boundary.parseBoundaryFromNullValue();

  public enum BoundType {
    INCLUSIVE,
    EXCLUSIVE;

    public BoundType getOp() {
      return this == INCLUSIVE ? EXCLUSIVE : INCLUSIVE;
    }

    public static BoundType getOp(BoundType boundType) {
      return boundType.getOp();
    }
  }

  private double lowerBound;
  private BoundType lowerBoundType;
  private double upperBound;
  private BoundType upperBoundType;
  private boolean integerValue;

  private Boundary() {
    super();
    this.integerValue = false;
  }

  public double getLowerBound() {
    return lowerBound;
  }

  public BoundType getLowerBoundType() {
    return lowerBoundType;
  }

  public double getUpperBound() {
    return upperBound;
  }

  public BoundType getUpperBoundType() {
    return upperBoundType;
  }

  private void setLowerBound(double lowerBound) {
    this.lowerBound = lowerBound;
  }

  private void setLowerBoundType(BoundType lowerBoundType) {
    this.lowerBoundType = lowerBoundType;
  }

  private void setUpperBound(double upperBound) {
    this.upperBound = upperBound;
  }

  private void setUpperBoundType(BoundType upperBoundType) {
    this.upperBoundType = upperBoundType;
  }

  public boolean isIntegerValue() {
    return integerValue;
  }

  private void setIntegerValue(boolean integerValue) {
    this.integerValue = integerValue;
  }

  /**
   * Check, if a number is in between the current boundary.
   *
   * @return true or false
   */
  public boolean isNumberInRange(double number) {
    if (number < lowerBound && lowerBound != NEGATIVE_INFINITY
        || number == NEGATIVE_INFINITY && lowerBound != NEGATIVE_INFINITY) {
      return false;
    }
    if (number > upperBound) {
      return false;
    }
    if (number == lowerBound && lowerBoundType == EXCLUSIVE) {
      return false;
    }
    if (number == upperBound && upperBoundType == EXCLUSIVE) {
      return false;
    }
    return true;
  }

  /**
   * Check, if a number is in between the current boundary.
   *
   * @return true or false
   */
  public boolean isNumberInRange(double number, BoundType boundType) {
    if ((number < lowerBound && lowerBound != NEGATIVE_INFINITY || number > upperBound)) {
      return false;
    } else if (boundType == INCLUSIVE
        || ((number > lowerBound || lowerBound == NEGATIVE_INFINITY) && number < upperBound)) {
      return isNumberInRange(number);
    } else {
      return false;
    }
  }

  /**
   * Check, if a other Boundary is overlapping (NOT equals and NO subsumption).
   *
   * <p>Example:
   *
   * <pre>
   * <blockquote>
   *
   *    true: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    false: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isOverlapping(Boundary other) {
    return !isEqualsTo(other) && !isNotInContact(other) && !isSubsumption(other);
  }

  /**
   * Check, if a other Boundary is a subsumption (NOT equals and NO overlapping) True, if this
   * subsumes other or other subsumes this.
   *
   * <p>Example:
   *
   * <pre>
   * <blockquote>
   *
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    true: (subsumption)
   *      o--o
   *       o-o
   *    false: (equals)
   *      o---o
   *      o---o
   *    true: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isSubsumption(Boundary other) {
    if (equals(other)) {
      return false;
    }

    if (other.lowerBound == this.lowerBound
            && (other.lowerBoundType == EXCLUSIVE && this.lowerBoundType == EXCLUSIVE)
        || (this.upperBound == other.upperBound
            && (this.upperBoundType == EXCLUSIVE && other.upperBoundType == EXCLUSIVE))) {
      return true;
    }

    if (this.lowerBound == other.lowerBound) {
      if (this.lowerBoundType == EXCLUSIVE) {
        return other.isNumberInRange(upperBound);
      }
      if (other.lowerBoundType == EXCLUSIVE) {
        return isNumberInRange(other.upperBound);
      }
    }

    if (this.upperBound == other.upperBound) {
      if (this.upperBoundType == EXCLUSIVE) {
        return other.isNumberInRange(lowerBound);
      }
      if (other.upperBoundType == EXCLUSIVE) {
        return isNumberInRange(other.lowerBound);
      }
    }

    return isNumberInRange(other.getLowerBound(), other.getLowerBoundType())
            && isNumberInRange(other.getUpperBound(), other.getUpperBoundType())
        || other.isNumberInRange(this.getLowerBound(), this.getLowerBoundType())
            && other.isNumberInRange(this.getUpperBound(), this.getUpperBoundType());
  }

  /**
   * Check if the current boundary subsumes the other boundary.
   *
   * @param other boundary
   * @return true, if the current boundary subsumes the other boundary
   */
  public boolean subsumes(Boundary other) {
    if (isSubsumption(other)) {
      return isNumberInRange(other.getLowerBound(), other.getLowerBoundType())
              && isNumberInRange(other.getUpperBound(), other.getUpperBoundType())
          || lowerBound == other.lowerBound
              && (lowerBoundType == INCLUSIVE || lowerBoundType == other.lowerBoundType)
              && upperBound == other.upperBound
              && (upperBoundType == INCLUSIVE || upperBoundType == other.upperBoundType);
    }
    return false;
  }

  /**
   * Check, if a other Boundary is NOT equals and NOT overlapping and NO subsumption.
   *
   * <p>Example:
   *
   * <pre>
   * <blockquote>
   *
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    true: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    false: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isNotInContact(Boundary other) {
    Objects.requireNonNull(other);
    return !this.isNumberInRange(other.getLowerBound(), other.getLowerBoundType())
        && !this.isNumberInRange(other.getUpperBound(), other.getUpperBoundType())
        && !other.isNumberInRange(this.getLowerBound(), this.getLowerBoundType())
        && !other.isNumberInRange(this.getUpperBound(), this.getUpperBoundType());
  }

  /**
   * Check, if a other Boundary is equals or overlapping or subsumption.
   *
   * <p>Example:
   *
   * <pre>
   * <blockquote>
   *
   *    true: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    true: (subsumption)
   *      o--o
   *       o-o
   *    true: (equals)
   *      o---o
   *      o---o
   *    true: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isInContact(Boundary other) {
    Objects.requireNonNull(other);
    return !isNotInContact(other);
  }

  /**
   * Check, if a other Boundary has equal bounds.
   *
   * <p>Example:
   *
   * <pre>
   * <blockquote>
   *
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    true: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  public boolean isEqualsTo(Boundary other) {
    return equals(other);
  }

  /**
   * Check, if a other Boundary has equal bounds.
   *
   * <p>Example:
   *
   * <pre>
   * <blockquote>
   *
   *    false: (overlapping)
   *     o------o
   *        o-----o
   *    false: (not overlapping)
   *     o---o
   *          o--o
   *    false: (subsumption)
   *      o--o
   *       o-o
   *    true: (equals)
   *      o---o
   *      o---o
   *    false: (subsumption)
   *      o------o
   *        o--o
   *  </blockquote></pre>
   *
   * @param other Boundary to check
   * @return true or false
   */
  @Override
  public boolean equals(Object other) {
    if (this == other) {
      return true;
    }
    if (other == null || getClass() != other.getClass()) {
      return false;
    }
    Boundary boundary = (Boundary) other;
    return Double.compare(boundary.lowerBound, lowerBound) == 0
        && Double.compare(boundary.upperBound, upperBound) == 0
        && lowerBoundType == boundary.lowerBoundType
        && upperBoundType == boundary.upperBoundType;
  }

  /**
   * Create a new Boundary, which is an intersection of the two given ones.
   *
   * <pre>
   * <blockquote>
   *        b1 : o----o
   *        b2 :   o----o
   *        new:   o--o
   *
   *        b1 : o----o
   *        b2 :   o-o
   *        new:   o-o:
   *
   *        b1 :   o----o
   *        b2 :  o--o
   *        new:   o-o
   *
   *        b1 :  o--o
   *        b2 : o----o
   *        new:  o--o
   * </blockquote></pre>
   *
   * @return a new Boundary as intersection
   */
  public Optional<Boundary> createBoundaryIntersection(Boundary other) {
    if (isNotInContact(other)) {
      return Optional.empty();
    }
    int compLower = compareToLowerBound(other);
    int compUpper = compareToUpperBound(other);
    return Optional.of(
        getBuilder()
            .withLowerBound(compLower > 0 ? getLowerBound() : other.getLowerBound())
            .withLowerBoundType(compLower > 0 ? getLowerBoundType() : other.getLowerBoundType())
            .withUpperBound(compUpper < 0 ? getUpperBound() : other.getUpperBound())
            .withUpperBoundType(compUpper < 0 ? getUpperBoundType() : other.getUpperBoundType())
            .isInteger(isIntegerValue())
            .build());
  }

  /**
   * Create a new Boundary with the lower bounds.
   *
   * <pre>
   * <blockquote>
   *
   *        b1 : o----o   (this)
   *        b2 :   o----o (other)
   *        new: o-o
   *
   *        b1 : o----o   (this)
   *        b2 :   o-o    (other)
   *        new: o-o
   *
   *        b1 :   o----o (this)
   *        b2 :  o--o    (other)
   *        new: empty Optional
   *
   *        b1 :  o--o    (this)
   *        b2 : o----o   (other)
   *        new: empty Optional
   * </blockquote></pre>
   *
   * @param other boundary
   * @return an Optional of a new Boundary
   */
  public Optional<Boundary> createBoundaryWithLowerBounds(Boundary other) {
    if (isIntegerValue()) {
      if (this.lowerBound + 1 == other.lowerBound
          && this.lowerBoundType == EXCLUSIVE
          && other.lowerBoundType == INCLUSIVE) {
        return Optional.empty();
      }
    }
    try {
      Boundary b =
          Boundary.getBuilder()
              .withLowerBound(this.getLowerBound())
              .withLowerBoundType(this.getLowerBoundType())
              .withUpperBound(other.getLowerBound())
              .withUpperBoundType(other.getLowerBoundType().getOp())
              .isInteger(other.isIntegerValue())
              .build();
      b.simplifyBound();
      return Optional.of(b);
    } catch (Exception e) {
      return Optional.empty();
    }
  }

  /**
   * Create a new Boundary with the upper bounds.
   *
   * <pre>
   * <blockquote>
   *
   *        b1 : o----o   (this)
   *        b2 :   o----o (other)
   *        new: empty Optional
   *
   *        b1 : o----o   (this)
   *        b2 :   o-o    (other)
   *        new:     oo
   *
   *        b1 :   o----o (this)
   *        b2 :  o--o    (other)
   *        new:     o--o
   *
   *        b1 :  o--o    (this)
   *        b2 : o----o   (other)
   *        new: empty Optional
   * </blockquote></pre>
   *
   * @param other boundary
   * @return an Optional of a new Boundary
   */
  public Optional<Boundary> createBoundaryWithUpperBounds(Boundary other) {
    if (isIntegerValue()) {
      if (this.upperBound == other.lowerBound + 1
          && this.upperBoundType == INCLUSIVE
          && other.upperBoundType == EXCLUSIVE) {
        return Optional.empty();
      }
    }
    try {
      Boundary b =
          Boundary.getBuilder()
              .withLowerBound(other.getUpperBound())
              .withLowerBoundType(other.getUpperBoundType().getOp())
              .withUpperBound(this.getUpperBound())
              .withUpperBoundType(this.getUpperBoundType())
              .isInteger(other.isIntegerValue())
              .build();
      b.simplifyBound();
      return Optional.of(b);
    } catch (Exception e) {
      return Optional.empty();
    }
  }

  private void simplifyBound() {
    if (isIntegerValue() && lowerBound + 1 == upperBound && lowerBoundType != upperBoundType) {
      if (upperBoundType == INCLUSIVE) {
        lowerBoundType = INCLUSIVE;
        lowerBound = upperBound;
      } else if (lowerBoundType == INCLUSIVE) {
        upperBoundType = INCLUSIVE;
        upperBound = lowerBound;
      }
    }
    if (isIntegerValue()
        && lowerBound + 1 == upperBound - 1
        && lowerBoundType == EXCLUSIVE
        && upperBoundType == EXCLUSIVE) {
      lowerBound = lowerBound + 1;
      upperBound = upperBound - 1;
      lowerBoundType = INCLUSIVE;
      upperBoundType = INCLUSIVE;
    }
  }

  /**
   * Create an new boundary, which is lover than the current. <br>
   * >10 --> <=10
   *
   * @return An Optional of the Boundary, if the current has a possible boundary, which is lower
   */
  public Optional<Boundary> createBoundaryLower() {
    if (lowerBound == NEGATIVE_INFINITY) {
      return Optional.empty();
    }
    return Optional.of(
        getBuilder()
            .withLowerBound(NEGATIVE_INFINITY)
            .withLowerBoundType(INCLUSIVE)
            .withUpperBound(lowerBound)
            .withUpperBoundType(lowerBoundType.getOp())
            .isInteger(isIntegerValue())
            .build());
  }

  /**
   * Create a new boundary between two existing ones.<br>
   * <10 & >20 --> [10..20]
   *
   * @param other {@link Boundary}
   * @return An Optional of the new Boundary between the two given ones Return an empty Optional, if
   *     there is no interval in between
   */
  public Optional<Boundary> createBoundaryBetween(Boundary other) {
    if (isInContact(Objects.requireNonNull(other))) {
      return Optional.empty();
    }
    // not space between the two bounds
    if (this.upperBound == other.lowerBound && this.upperBoundType != other.lowerBoundType
        || other.upperBound == this.lowerBound && other.upperBoundType != this.lowerBoundType) {
      return Optional.empty();
    }
    // check if is integer val
    if (isIntegerValue()) {
      if (this.upperBound + 1 == other.lowerBound
              && this.upperBoundType == INCLUSIVE
              && other.lowerBoundType == INCLUSIVE
          || other.upperBound + 1 == this.lowerBound
              && other.upperBoundType == INCLUSIVE
              && this.lowerBoundType == INCLUSIVE) {
        return Optional.empty();
      }
    }
    if (this.upperBound < other.upperBound) {
      return Optional.of(
          getBuilder()
              .withLowerBound(this.upperBound)
              .withLowerBoundType(this.upperBoundType.getOp())
              .withUpperBound(other.lowerBound)
              .withUpperBoundType(other.lowerBoundType.getOp())
              .isInteger(isIntegerValue())
              .build());
    } else {
      return Optional.of(
          getBuilder()
              .withLowerBound(other.upperBound)
              .withLowerBoundType(other.upperBoundType.getOp())
              .withUpperBound(this.lowerBound)
              .withUpperBoundType(this.lowerBoundType.getOp())
              .isInteger(isIntegerValue())
              .build());
    }
  }

  /**
   * Create a new boundary, which begins at the upper bound of the original. <br>
   * <10 --> >=10
   *
   * @return An Optional of the Boundary if the current has a possible boundary, which is higher
   */
  public Optional<Boundary> createBoundaryUpper() {
    if (upperBound == POSITIVE_INFINITY) {
      return Optional.empty();
    }
    return Optional.of(
        getBuilder()
            .withLowerBound(upperBound)
            .withLowerBoundType(upperBoundType.getOp())
            .withUpperBound(POSITIVE_INFINITY)
            .withUpperBoundType(INCLUSIVE)
            .isInteger(isIntegerValue())
            .build());
  }

  /**
   * Create a new Boundary from two given ones. The Optional is empty, if the given two bounds are
   * not direct followed by each other or current in contact<u>Example:</u><br>
   * <code>
   * [10..20]  &  >20 --> >=10<br> [10..20]  &  <10 --> <=20<br> [10..20]  &  <=0 --> empty Optional
   * (Boundaries are not in contact)<br> [10..20]  &  <=10 --> empty Optional (Boundaries are
   * overlapping)<br>
   * </code>
   *
   * @param other the other boundary
   */
  public Optional<Boundary> appendWith(Boundary other) {
    if (isNotInContact(other)) {
      if (this.upperBound == other.lowerBound && this.upperBoundType != other.lowerBoundType
          || isIntegerValue()
              && this.upperBound + 1 == other.lowerBound
              && this.upperBoundType == INCLUSIVE
              && other.lowerBoundType == INCLUSIVE) {
        return Optional.of(
            getBuilder()
                .withLowerBound(this.lowerBound)
                .withLowerBoundType(this.lowerBoundType)
                .withUpperBound(other.upperBound)
                .withUpperBoundType(other.getUpperBoundType())
                .isInteger(isIntegerValue())
                .build());
      } else if (this.lowerBound == other.upperBound && other.upperBoundType != this.lowerBoundType
          || isIntegerValue()
              && other.upperBound + 1 == this.lowerBound
              && other.upperBoundType == INCLUSIVE
              && this.lowerBoundType == INCLUSIVE) {
        return Optional.of(
            getBuilder()
                .withLowerBound(other.lowerBound)
                .withLowerBoundType(other.lowerBoundType)
                .withUpperBound(this.upperBound)
                .withUpperBoundType(this.getUpperBoundType())
                .isInteger(isIntegerValue())
                .build());
      }
    }
    return Optional.empty();
  }

  /**
   * Create a new Boundary from two given ones. The Optional is empty, if the given two bounds are
   * not direct followed by each other. They can be in contact <u>Example:</u><br>
   * <code>
   * [10..20]  &  >20 --> >=10<br> [10..20]  &  <10 --> <=20<br> [10..20]  &  <=0 --> empty Optional
   * (Boundaries are not in contact)<br> [10..20]  &  <=10 --> <=20 <br>
   * </code>
   *
   * @param other the other boundary
   */
  public Optional<Boundary> combineWith(Boundary other) {
    if (isNotInContact(other)) {
      return appendWith(other);
    }
    Builder builder = getBuilder().isInteger(isIntegerValue());
    if (this.lowerBound == NEGATIVE_INFINITY || this.lowerBound < other.lowerBound) {
      builder.withLowerBound(this.lowerBound).withLowerBoundType(this.lowerBoundType);
    } else if (other.lowerBound == NEGATIVE_INFINITY || other.lowerBound < this.lowerBound) {
      builder.withLowerBound(other.lowerBound).withLowerBoundType(other.lowerBoundType);
    } else {
      builder
          .withLowerBound(other.lowerBound)
          .withLowerBoundType(
              this.lowerBoundType == INCLUSIVE || other.lowerBoundType == INCLUSIVE
                  ? INCLUSIVE
                  : EXCLUSIVE);
    }
    if (this.upperBound == POSITIVE_INFINITY || this.upperBound > other.upperBound) {
      builder.withUpperBound(this.upperBound).withUpperBoundType(this.upperBoundType);
    } else if (other.upperBound == POSITIVE_INFINITY || other.upperBound > this.upperBound) {
      builder.withUpperBound(other.upperBound).withUpperBoundType(other.upperBoundType);
    } else {
      builder
          .withUpperBound(other.upperBound)
          .withUpperBoundType(
              this.upperBoundType == INCLUSIVE || other.upperBoundType == INCLUSIVE
                  ? INCLUSIVE
                  : EXCLUSIVE);
    }
    return Optional.of(builder.build());
  }

  @Override
  public int hashCode() {
    return Objects.hash(lowerBound, lowerBoundType, upperBound, upperBoundType);
  }

  /**
   * Parse a String to a boundary object.<br>
   * <u>Possible forms:</u><br>
   * <code>
   * Interval: [x..y] (inclusive x and inclusive y) <br> Interval: ]x..y[ (exclusive x and exclusive
   * y) <br> Max Val: <=x (inclusive x) <br> Max Val: &lt;x (exclusive x) <br> Min Val: >=x
   * (inclusive x) <br> Min Val: >x (exclusive x) <br>Equal: =x (inclusive x) <br>
   * </code>
   *
   * @param expr String to parse
   * @return a new Boundary object
   */
  public static Boundary parseFromNumericValue(String expr) {
    return parseFromNumericValue(expr, false);
  }

  /**
   * Parse a String to a boundary object.<br>
   * <u>Possible forms:</u><br>
   * <code>
   * Interval: [x..y] (inclusive x and inclusive y) <br> Interval: ]x..y[ (exclusive x and exclusive
   * y) <br> Max Val: <=x (inclusive x) <br> Max Val: &lt;x (exclusive x) <br> Min Val: >=x
   * (inclusive x) <br> Min Val: >x (exclusive x) <br>Equal: =x (inclusive x) <br>
   * </code>
   *
   * @param expr String to parse
   * @param isIntegerValue true, if the value is a integer Value
   * @return a new Boundary object
   */
  public static Boundary parseFromNumericValue(String expr, boolean isIntegerValue) {
    Builder builder = getBuilder();
    if (isIntegerValue) {
      builder.isInteger();
    }
    try {
      expr = Objects.requireNonNull(expr).trim();
      if (expr.startsWith("[") || expr.startsWith("]")) {
        builder
            .withLowerBoundType(expr.startsWith("[") ? INCLUSIVE : EXCLUSIVE)
            .withUpperBoundType(expr.endsWith("]") ? INCLUSIVE : EXCLUSIVE);
        expr = expr.replace('[', '\0').replace(']', '\0');
        String[] n = expr.split("\\.\\.");
        builder.withLowerBound(Double.parseDouble(n[0])).withUpperBound(Double.parseDouble(n[1]));
      } else if (expr.startsWith(">")) {
        builder.withLowerBoundType(expr.contains("=") ? INCLUSIVE : EXCLUSIVE);
        expr = expr.replace('=', '\0').replace('>', '\0');
        builder
            .withLowerBound(Double.parseDouble(expr))
            .withUpperBound(POSITIVE_INFINITY)
            .withUpperBoundType(INCLUSIVE);
      } else if (expr.startsWith("<")) {
        builder.withUpperBoundType(expr.contains("=") ? INCLUSIVE : EXCLUSIVE);
        expr = expr.replace('=', '\0').replace('<', '\0');
        builder
            .withUpperBound(Double.parseDouble(expr))
            .withLowerBound(NEGATIVE_INFINITY)
            .withLowerBoundType(INCLUSIVE);
      } else if (expr.startsWith("=")) {
        expr = expr.replace('=', '\0');
        builder
            .withUpperBound(Double.parseDouble(expr))
            .withLowerBound(Double.parseDouble(expr))
            .withLowerBoundType(INCLUSIVE)
            .withUpperBoundType(INCLUSIVE);
      } else {
        throw new IllegalArgumentException();
      }
    } catch (Exception e) {
      throw new IllegalArgumentException(
          String.format("Parsing boundary failed: %s. Exception: %s", expr, e.getMessage()));
    }
    return builder.build();
  }

  /**
   * Parse a String value to a boundary object. The Value of the String equals the hashcode of the
   * string value and has the form<br>
   * <code>={stringValue.hashCode()}</code>
   *
   * @param stringValue the string to create the boundary
   * @return a boundary
   */
  public static Boundary parseFromStringValue(String stringValue) {
    int hash = Objects.requireNonNull(stringValue).hashCode();
    return getBuilder()
        .withUpperBound(hash)
        .withLowerBound(hash)
        .withLowerBoundType(INCLUSIVE)
        .withUpperBoundType(INCLUSIVE)
        .build();
  }

  /**
   * Parse a String value to a boundary object. The Value of the String equals the hashcode of the
   * string value and has the form<br>
   * <code>={stringValue.hashCode()}</code>
   *
   * @return a boundary
   */
  public static Boundary parseFromBooleanValue(boolean bolVal) {
    int val = bolVal ? 0 : 1;
    return getBuilder()
        .withUpperBound(val)
        .withLowerBound(val)
        .withLowerBoundType(INCLUSIVE)
        .withUpperBoundType(INCLUSIVE)
        .build();
  }

  /**
   * Creates a "wildcard" boundary. [-infinity..+infinity]
   *
   * @return a boundary
   */
  public static Boundary parseBoundaryFromNullValue() {
    // Wildcard
    return getBuilder()
        .withLowerBound(NEGATIVE_INFINITY)
        .withUpperBound(POSITIVE_INFINITY)
        .withLowerBoundType(INCLUSIVE)
        .withUpperBoundType(INCLUSIVE)
        .build();
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    if (getLowerBound() == NEGATIVE_INFINITY && getUpperBound() == POSITIVE_INFINITY) {
      // Wildcard
      sb.append("-");
    } else if (getLowerBound() == NEGATIVE_INFINITY) {
      sb.append("<");
      if (getUpperBoundType() == INCLUSIVE) {
        sb.append("=");
      }
      sb.append(getValAsString(getUpperBound()));
    } else if (getUpperBound() == POSITIVE_INFINITY) {
      sb.append(">");
      if (getLowerBoundType() == INCLUSIVE) {
        sb.append("=");
      }
      sb.append(getValAsString(getLowerBound()));
    } else if (getLowerBound() == getUpperBound()) {
      sb.append("=").append(getValAsString(getLowerBound()));
    } else {
      if (getLowerBoundType() == INCLUSIVE) {
        sb.append("[");
      } else {
        sb.append("]");
      }
      sb.append(getValAsString(getLowerBound()));
      sb.append("..");
      sb.append(getValAsString(getUpperBound()));
      if (getUpperBoundType() == INCLUSIVE) {
        sb.append("]");
      } else {
        sb.append("[");
      }
    }
    return sb.toString();
  }

  private String getValAsString(double value) {
    if (isIntegerValue()) {
      return String.valueOf((new Double(value)).intValue());
    }
    return String.valueOf((new Double(value)));
  }

  @Override
  public int compareTo(Boundary o) {
    if (o == null) {
      return 1;
    }
    int comLower = compareToLowerBound(o);
    return comLower == 0 ? compareToUpperBound(o) : comLower;
  }

  /**
   * Compare the lower bounds.
   *
   * @param o other boundary
   * @return -1 if current lower bound is lower, 0 if equal, 1 if higher
   */
  public int compareToLowerBound(Boundary o) {
    if (lowerBound == o.lowerBound) {
      if (lowerBoundType == o.lowerBoundType) {
        return 0;
      }
      return lowerBoundType == INCLUSIVE ? -1 : 1;
    } else if (lowerBound == NEGATIVE_INFINITY) {
      return -1;
    } else if (o.lowerBound == NEGATIVE_INFINITY) {
      return 1;
    } else {
      return Double.compare(lowerBound, o.lowerBound);
    }
  }

  /**
   * Compare the upper bounds.
   *
   * @param o other boundary
   * @return -1 if current upper bound is lower, 0 if equal, 1 if higher
   */
  public int compareToUpperBound(Boundary o) {
    if (upperBound == o.upperBound) {
      if (upperBoundType == o.upperBoundType) {
        return 0;
      }
      return upperBoundType == INCLUSIVE ? 1 : -1;
    } else if (upperBound == POSITIVE_INFINITY) {
      return 1;
    } else if (o.upperBound == POSITIVE_INFINITY) {
      return -1;
    } else {
      return Double.compare(upperBound, o.upperBound);
    }
  }

  public boolean isInfite() {
    return lowerBound == NEGATIVE_INFINITY && upperBound == POSITIVE_INFINITY;
  }

  @Override
  public Object clone() throws CloneNotSupportedException {
    return getBuilder()
        .withLowerBound(lowerBound)
        .withLowerBoundType(lowerBoundType)
        .withUpperBound(upperBound)
        .withUpperBoundType(upperBoundType)
        .isInteger(integerValue)
        .build();
  }

  public static Builder getBuilder() {
    return new Boundary().new Builder();
  }

  public class Builder extends DefaultBuilder<Boundary> {

    public Builder withUpperBound(double bound) {
      value.setUpperBound(bound);
      return this;
    }

    public Builder withLowerBound(double bound) {
      value.setLowerBound(bound);
      return this;
    }

    public Builder withUpperBoundType(BoundType boundType) {
      value.setUpperBoundType(boundType);
      return this;
    }

    public Builder withLowerBoundType(BoundType boundType) {
      value.setLowerBoundType(boundType);
      return this;
    }

    public Builder isInteger() {
      return isInteger(true);
    }

    public Builder isInteger(boolean val) {
      value.setIntegerValue(val);
      return this;
    }

    @Override
    protected void validate() {
      if (value.lowerBound == NEGATIVE_INFINITY) {
        value.lowerBoundType = INCLUSIVE;
      }
      if (value.upperBound == POSITIVE_INFINITY) {
        value.upperBoundType = INCLUSIVE;
      }
      Objects.requireNonNull(value.lowerBoundType, "Lower bound type is null");
      Objects.requireNonNull(value.upperBoundType, "Upper bound type is null");
      if (value.lowerBound > value.upperBound && NEGATIVE_INFINITY != value.lowerBound) {
        throw new IllegalArgumentException(
            String.format(
                "Lower bound %f is greater than upper bound %f.",
                value.lowerBound, value.upperBound));
      }
      // if equal -> both must be INCLUSIVE!
      if (value.upperBound == value.lowerBound
          && (value.lowerBoundType != INCLUSIVE || value.upperBoundType != INCLUSIVE)) {
        throw new IllegalArgumentException("If bounds are equal -> both bounds must be INCLUSIVE!");
      }
      // if boundary values are declared as integer, than there should be no not int bounds
      if (value.isIntegerValue()) {
        if (value.lowerBound != NEGATIVE_INFINITY
            && ((value.lowerBound < 0 && value.lowerBound != Math.ceil(value.lowerBound))
                || value.lowerBound != Math.floor(value.lowerBound))) {
          throw new IllegalArgumentException(
              String.format("Lower bound %f is no integer/long value.", value.lowerBound));
        }
        if (value.upperBound != POSITIVE_INFINITY
            && ((value.upperBound < 0 && value.upperBound != Math.ceil(value.upperBound))
                || value.upperBound != Math.floor(value.upperBound))) {
          throw new IllegalArgumentException(
              String.format("Upper bound %f is no integer/long value.", value.upperBound));
        }
      }
    }
  }
}
