package de.unikoblenz.fgbks.dmn.core.dmnhelper;

import de.unikoblenz.fgbks.dmn.core.models.RuleIdentifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import org.camunda.bpm.dmn.engine.DmnDecision;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableImpl;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableInputImpl;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableOutputImpl;
import org.camunda.bpm.dmn.engine.impl.DmnDecisionTableRuleImpl;
import org.camunda.bpm.dmn.engine.impl.DmnExpressionImpl;

/**
 * Class for storing inputs and outputs from DMNDecisions. Also this class calculate multi tables
 * with same output columns to calculate verifications across tables.
 */
public class DmnRuleMap {

  private List<DmnDecision> dmnDecisions;
  private Map<InputType, List<Value>> inputs;
  private List<InputType> inputColumns;
  private Map<OutputType, List<Value>> outputs;
  private List<OutputType> outputColumns;
  private Set<String> dmnDecisionsKeys;
  private Set<String> mergedDecisionsKeys;
  private Map<RuleIdentifier, List<Value>> ruleInputValues;
  private Map<RuleIdentifier, List<Value>> ruleOutputValues;

  private DmnRuleMap(List<DmnDecision> dmnDecisions) {
    this.dmnDecisions = new ArrayList<>(dmnDecisions);
    this.inputs = new HashMap<>();
    this.inputColumns = new ArrayList<>();
    this.outputs = new HashMap<>();
    this.outputColumns = new ArrayList<>();
    this.dmnDecisionsKeys = new HashSet<>();
    this.mergedDecisionsKeys = new HashSet<>();
    this.ruleInputValues = new HashMap<>();
    this.ruleOutputValues = new HashMap<>();
  }

  /**
   * Create a Map for all inputs and all outputs out of a list of DMN Decisions
   *
   * @param decisionList List of DMNDecision @NotNull
   * @return a {@link DmnRuleMap} from the list of DMNDecision
   */
  public static DmnRuleMap createMapFromDmn(List<DmnDecision> decisionList) {
    DmnRuleMap dmnRuleMap = new DmnRuleMap(Objects.requireNonNull(decisionList));
    for (DmnDecision d : decisionList) {
      createMapFromDmn(dmnRuleMap, d);
    }
    // calculate new merged dmn decisions
    dmnRuleMap.calculateMergedTables();
    // return the map
    return dmnRuleMap;
  }

  /**
   * Create a Map for all inputs and all outputs out of a list of DMN Decisions
   *
   * @param d DMNDecision @NotNull
   * @return a {@link DmnRuleMap} from the list of DMNDecision
   */
  public static DmnRuleMap createMapFromDmn(DmnDecision d) {
    DmnRuleMap dmnRuleMap = new DmnRuleMap(Collections.singletonList(Objects.requireNonNull(d)));
    return createMapFromDmn(dmnRuleMap, d);
  }

  private static DmnRuleMap createMapFromDmn(DmnRuleMap dmnRuleMap, DmnDecision d) {
    if (d.isDecisionTable()) {
      dmnRuleMap.dmnDecisionsKeys.add(d.getKey());
      DmnDecisionTableImpl table = (DmnDecisionTableImpl) d.getDecisionLogic();
      // inputs
      int i = 0;
      for (DmnDecisionTableInputImpl tableInput : table.getInputs()) {
        InputType it = new InputType(d, tableInput);
        dmnRuleMap.inputColumns.add(it);
        dmnRuleMap.inputs.put(it, new ArrayList<>());
        List<Value> inList = dmnRuleMap.inputs.get(it);
        int r = 0;
        for (DmnDecisionTableRuleImpl rule : table.getRules()) {
          DmnExpressionImpl condition = rule.getConditions().get(i);
          Value value = it.getValue(condition, d, ++r, rule.getId());
          inList.add(value);
          List<Value> rules =
              dmnRuleMap.ruleInputValues.getOrDefault(value.getRuleIdentifier(), new ArrayList<>());
          rules.add(value);
          dmnRuleMap.ruleInputValues.put(value.getRuleIdentifier(), rules);
        }
        i++;
      }
      // outputs
      i = 0;
      for (DmnDecisionTableOutputImpl tableOutput : table.getOutputs()) {
        OutputType ot = new OutputType(d, tableOutput);
        dmnRuleMap.outputColumns.add(ot);
        dmnRuleMap.outputs.put(ot, new ArrayList<>());
        List<Value> outList = dmnRuleMap.outputs.get(ot);
        int r = 0;
        for (DmnDecisionTableRuleImpl rule : table.getRules()) {
          DmnExpressionImpl conclusion = rule.getConclusions().get(i);
          Value value = ot.getValue(conclusion, d, ++r, rule.getId());
          outList.add(value);
          List<Value> rules =
              dmnRuleMap.ruleOutputValues.getOrDefault(
                  value.getRuleIdentifier(), new ArrayList<>());
          rules.add(value);
          dmnRuleMap.ruleOutputValues.put(value.getRuleIdentifier(), rules);
        }
        i++;
      }
    }
    return dmnRuleMap;
  }

  private void calculateMergedTables() {
    // search identical outputs;
    Set<Set<OutputType>> identicalOutputs = new HashSet<>();
    for (OutputType ot : outputColumns) {
      Set<OutputType> equalOuts = new HashSet<>();
      for (OutputType otInner : outputColumns) {
        if (ot.label != null
            && ot.label.equals(otInner.label)
            && ot.dataType.equals(otInner.dataType)) {
          equalOuts.add(otInner);
        }
      }
      if (equalOuts.size() > 1) {
        identicalOutputs.add(equalOuts);
      }
    }
    // create new tables
    for (Set<OutputType> outputTypes : identicalOutputs) {
      createNewMergedTable(outputTypes);
    }
  }

  private void createNewMergedTable(Set<OutputType> outputTypes) {
    String newDMNKey = "MergedDecision_" + UUID.randomUUID().toString().substring(0, 5);
    mergedDecisionsKeys.add(newDMNKey);
    OutputType newOutputType = null;
    List<InputType> newInputTypes = new ArrayList<>();
    // Create new Output and Input Columns
    for (OutputType ot : outputTypes) {
      // create a new Output type
      if (newOutputType == null) {
        newOutputType = new OutputType(ot);
        newOutputType.setOverwrittenDecisionKey(newDMNKey);
        outputColumns.add(newOutputType);
      }
      // create new input types
      for (InputType inputType :
          inputColumns.stream()
              .filter(i -> i.getDecisionKey().equals(ot.getDecisionKey()))
              .collect(Collectors.toList())) {
        // only, if the label didnt exists in prev created cols
        if (newInputTypes.stream()
                .filter(i -> i.label.equals(inputType.label) && i.dataType == inputType.dataType)
                .limit(1)
                .count()
            < 1) {
          InputType newInputType = new InputType(inputType);
          newInputType.setOverwrittenDecisionKey(newDMNKey);
          newInputTypes.add(newInputType);
        }
      }
    }
    inputColumns.addAll(newInputTypes);
    // add the values to the new cols
    for (OutputType ot : outputTypes) {
      { // add output values
        List<Value> vals = outputs.getOrDefault(newOutputType, new ArrayList<>());
        vals.addAll(outputs.get(ot));
        outputs.put(newOutputType, vals);
      }
      int rowCount = outputs.get(ot).size();
      List<InputType> orgInputTypes =
          inputColumns.stream()
              .filter(i -> i.getDecisionKey().equals(ot.getDecisionKey()))
              .collect(Collectors.toList());
      // create new input values, of existing ones of current output col
      List<RuleIdentifier> ruleIdentifiers = new ArrayList<>();
      for (InputType inputType : newInputTypes) {
        // try to find existing col
        Optional<InputType> orgI =
            orgInputTypes.stream()
                .filter(i -> i.label.equals(inputType.label) && i.dataType == inputType.dataType)
                .findFirst();
        List<Value> vals = inputs.getOrDefault(inputType, new ArrayList<>());
        if (orgI.isPresent()) {
          // copy existing
          vals.addAll(inputs.get(orgI.get()));
          ruleIdentifiers =
              inputs.get(orgI.get()).stream()
                  .map(Value::getRuleIdentifier)
                  .collect(Collectors.toList());
          inputs.put(inputType, vals);
        }
      }
      // create new input values, of non existing ones of current output col
      for (InputType inputType : newInputTypes) {
        // try to find existing col
        Optional<InputType> orgI =
            orgInputTypes.stream()
                .filter(i -> i.label.equals(inputType.label) && i.dataType == inputType.dataType)
                .findFirst();
        List<Value> vals = inputs.getOrDefault(inputType, new ArrayList<>());
        if (!orgI.isPresent()) {
          // create new wildcards
          for (int i = 0; i < rowCount; i++) {
            vals.add(
                new Value(
                    Boundary.INFINITY_BOUNDARY,
                    null,
                    RuleIdentifier.getBuilder()
                        .withConditionName(null)
                        .withConditionId(null)
                        .withRowNumber(i + 1)
                        .withRuleId(ruleIdentifiers.get(i).getRuleId())
                        .withDecisionKey(ruleIdentifiers.get(i).getDecisionKey())
                        .withDecisionName(ruleIdentifiers.get(i).getTableName())
                        .build(),
                    inputType));
          }
          inputs.put(inputType, vals);
        }
      }
    }
  }

  public Set<String> getMergedDecisionsKeys() {
    return new HashSet<>(mergedDecisionsKeys);
  }

  /**
   * Get the all rules from the given dmn decision
   *
   * @param d the dmnDecision
   * @return A list of RuleIdentifier representing the rules
   */
  public List<RuleIdentifier> getRulesFromDMNTable(DmnDecision d) {
    return getRulesFromDMNTable(Objects.requireNonNull(d).getKey());
  }

  /**
   * Get the all rules from the given dmn decision
   *
   * @param dmnKey the dmnKey of the dmn decisions
   * @return A list of RuleIdentifier representing the rules
   */
  public List<RuleIdentifier> getRulesFromDMNTable(String dmnKey) {
    return ruleInputValues.keySet().stream()
        .filter(x -> x.getDecisionKey().equals(Objects.requireNonNull(dmnKey)))
        .sorted(Comparator.comparingInt(RuleIdentifier::getRowNumber))
        .collect(Collectors.toList());
  }

  /**
   * Get all input columns for the given RuleIdentifier
   *
   * @param ruleIdentifier the RuleIdentifier
   * @return a list of values (columns) of the current rule
   */
  public List<Value> getRuleInputValuesFromIdentifier(RuleIdentifier ruleIdentifier) {
    if (!ruleInputValues.containsKey(ruleIdentifier)) {
      throw new IllegalStateException(
          "RuleIdentifier not found in input List: " + ruleIdentifier.toString());
    }
    return ruleInputValues.get(ruleIdentifier);
  }

  /**
   * Return a list of all input values from a given type.
   *
   * @param type the given input type
   * @return a list of all Values from the type
   */
  public List<Value> getValuesFromInputType(Type type) {
    if (!inputs.containsKey(type)) {
      return Collections.emptyList();
    }
    return new ArrayList<>(inputs.get(type));
  }

  /**
   * Return a list of input values with the given type, filtered by the given rule identifiers
   *
   * @param type the given input type
   * @param ruleIdentifiersFilter the rules to filter
   * @return a list of given input types
   */
  public List<Value> getValuesFromInputType(Type type, List<RuleIdentifier> ruleIdentifiersFilter) {
    if (!inputs.containsKey(type)) {
      return Collections.emptyList();
    }
    if (ruleIdentifiersFilter == null) {
      return getValuesFromInputType(type);
    }
    return inputs.get(type).stream()
        .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
        .collect(Collectors.toList());
  }

  /**
   * Return a list of all output values from a given type.
   *
   * @param type the given output type
   * @return a list of all values from the given output type
   */
  public List<Value> getValuesFromOutputType(Type type) {
    if (!outputs.keySet().contains(type)) {
      return Collections.emptyList();
    }
    return new ArrayList<>(outputs.get(type));
  }

  /**
   * Return a list of output values with the given type, filtered by the given rule identifiers
   *
   * @param type the given output type
   * @param ruleIdentifiersFilter the rules to filter
   * @return a list of all values from the given output type
   */
  public List<Value> getValuesFromOutputType(
      Type type, List<RuleIdentifier> ruleIdentifiersFilter) {
    if (!outputs.keySet().contains(type)) {
      return Collections.emptyList();
    }
    if (ruleIdentifiersFilter == null) {
      return getValuesFromOutputType(type);
    }
    return outputs.get(type).stream()
        .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
        .collect(Collectors.toList());
  }

  /**
   * Return a list of output or input values with the given type
   *
   * @param type the given output or input type
   * @return a list of all values from the given output or input type
   */
  public List<Value> getValuesFromType(Type type) {
    if (outputs.keySet().contains(type)) {
      return new ArrayList<>(outputs.get(type));
    } else if (inputs.keySet().contains(type)) {
      return new ArrayList<>(inputs.get(type));
    }
    return Collections.emptyList();
  }

  /**
   * Return a list of output or input values with the given type, filtered by the given rule
   * identifiers
   *
   * @param type the given output or input type
   * @param ruleIdentifiersFilter the rules to filter
   * @return a list of all values from the given output or input type
   */
  public List<Value> getValuesFromType(Type type, List<RuleIdentifier> ruleIdentifiersFilter) {
    if (ruleIdentifiersFilter == null) {
      return getValuesFromType(type);
    }
    if (outputs.keySet().contains(type)) {
      return outputs.get(type).stream()
          .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
          .collect(Collectors.toList());
    } else if (inputs.keySet().contains(type)) {
      return inputs.get(type).stream()
          .filter(r -> ruleIdentifiersFilter.contains(r.getRuleIdentifier()))
          .collect(Collectors.toList());
    }
    return Collections.emptyList();
  }

  /**
   * Return a list of input types, which belongs to a given decisionKey
   *
   * @param decisionKey the decisionKey
   * @return a list of Input types
   */
  public List<Type> getAllInputTypesFromDecisionKey(String decisionKey) {
    return inputColumns.stream()
        .filter(i -> i.getDecisionKey().equals(Objects.requireNonNull(decisionKey)))
        .collect(Collectors.toList());
  }

  /**
   * Return a list of output types, which belongs to a given decisionKey
   *
   * @param decisionKey the decisionKey
   * @return a list of Output types
   */
  public List<Type> getAllOutputTypesFromDecisionKey(String decisionKey) {
    return outputColumns.stream()
        .filter(i -> i.getDecisionKey().equals(decisionKey))
        .collect(Collectors.toList());
  }

  /**
   * Return a Set of input and output types, which belongs to a given decisionKey
   *
   * @param decisionKey the decisionKey
   * @return a set of output and input types
   */
  public List<Type> getAllTypesFromDecisionKey(String decisionKey) {
    List<Type> types = getAllInputTypesFromDecisionKey(decisionKey);
    types.addAll(getAllOutputTypesFromDecisionKey(decisionKey));
    return types;
  }
}
