package de.unikoblenz.fgbks.dmn.core.models;

import de.unikoblenz.fgbks.dmn.core.dmnhelper.DmnRuleMap;
import de.unikoblenz.fgbks.dmn.core.models.VerifierCollectionResult.Builder;
import de.unikoblenz.fgbks.dmn.core.verifier.*;
import de.unikoblenz.fgbks.dmn.core.verifier.multitable.MultiTableEquivalentVerifier;
import de.unikoblenz.fgbks.dmn.core.verifier.multitable.MultiTableIdenticalVerifier;
import de.unikoblenz.fgbks.dmn.core.verifier.multitable.MultiTableOverlappingVerifier;
import de.unikoblenz.fgbks.dmn.core.verifier.multitable.MultiTablePartialReductionVerifier;
import de.unikoblenz.fgbks.dmn.core.verifier.multitable.MultiTableSubsumptionVerifier;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ConcurrentLinkedQueue;
import org.camunda.bpm.dmn.engine.DmnDecision;
import org.slf4j.LoggerFactory;

public enum VerifierType {
  Identical("Checking for identical rules.", IdenticalVerifier.class, true, true, false),
  Subsumption(
      "Checking for rules, which subsume other rules.",
      SubsumptionVerifier.class,
      true,
      true,
      false),
  Equivalent("Checking for synonyms in columns.", EquivalentVerifier.class, true, false, false),
  Overlap("Checking for overlapping rules.", OverlappingVerifier.class, true, true, false),
  Missing("Checking for missing rules.", MissingVerifier.class, true, false, false),
  PartialReduction(
      "Checking for partial reduction of rules (combination).",
      PartialReductionVerifier.class,
      true,
      false,
      false),
  Interdeterminism(
      "Identical, Overlapping or Subsumption with different conclusions.",
      InterdeterminismVerifier.class,
      false,
      false,
      false),
  MultiTableIdentical(
      "Checking for identical rules in multiple tables with identical outcome column.",
      MultiTableIdenticalVerifier.class,
      true,
      true,
      true),
  MultiTableOverlapping(
      "Checking for overlapping rules in multiple tables with identical outcome column.",
      MultiTableOverlappingVerifier.class,
      true,
      true,
      true),
  MultiTableSubsumption(
      "Checking for subsumptions in multiple tables with identical outcome column.",
      MultiTableSubsumptionVerifier.class,
      true,
      true,
      true),
  MultiTableEquivalent(
      "Checking for synonyms in columns in multiple tables with identical outcome column.",
      MultiTableEquivalentVerifier.class,
      true,
      false,
      true),
  MultiTablePartialReduction(
      "Checking for partial reduction of rules (combination) in multiple tables with identical outcome column.",
      MultiTablePartialReductionVerifier.class,
      true,
      false,
      true);

  private final Class<? extends AbstractVerifier> verifierClass;
  private String description;
  private boolean autoExecutable;
  private boolean hasInterdeterminismRules;
  private boolean isMultiTable;

  VerifierType(
      String description,
      Class<? extends AbstractVerifier> verifierClass,
      boolean autoExecutable,
      boolean hasInterdeterminismRules,
      boolean isMultiTable) {
    this.description = description;
    this.verifierClass = verifierClass;
    this.autoExecutable = autoExecutable;
    this.hasInterdeterminismRules = hasInterdeterminismRules;
    this.isMultiTable = isMultiTable;
  }

  public String getDescription() {
    return description;
  }

  private boolean isAutoExecutable() {
    return autoExecutable;
  }

  private boolean hasInterdeterminismRules() {
    return hasInterdeterminismRules;
  }

  public boolean isMultiTable() {
    return isMultiTable;
  }

  public Optional<VerifierResult> getResult(List<DmnDecision> dmnDecisionList) {
    return getResult(dmnDecisionList, null, null);
  }

  private Optional<VerifierResult> getResult(
      List<DmnDecision> dmnDecisionList,
      DmnRuleMap dmnRuleMap,
      ConcurrentLinkedQueue<VerificationResult> interdeterminismRules) {
    if (dmnRuleMap == null) {
      dmnRuleMap = DmnRuleMap.createMapFromDmn(dmnDecisionList);
    }
    try {
      AbstractVerifier verifier = verifierClass.newInstance().withVerifierType(this);
      if (!isAutoExecutable() && interdeterminismRules != null) {
        verifier.setInterdeterminismRules(interdeterminismRules);
      }
      VerifierResult result = verifier.findValidationErrors(dmnDecisionList, dmnRuleMap);
      if (hasInterdeterminismRules()) {
        interdeterminismRules.addAll(verifier.getInterdeterminismRules());
      }
      return Optional.of(result);
    } catch (Exception ex) {
      LoggerFactory.getLogger(VerifierType.class).warn(ex.getMessage());
      ex.printStackTrace();
    }
    return Optional.empty();
  }

  public static VerifierCollectionResult getAllResults(List<DmnDecision> dmnDecisionList) {
    return getAllResults(dmnDecisionList, false);
  }

  public static VerifierCollectionResult getAllResults(
      List<DmnDecision> dmnDecisionList, boolean includeMultiTable) {
    DmnRuleMap dmnRuleMap = DmnRuleMap.createMapFromDmn(dmnDecisionList);
    ConcurrentLinkedQueue<VerificationResult> interdeterminismRules = new ConcurrentLinkedQueue<>();
    Builder builder = VerifierCollectionResult.getBuilder();
    Arrays.asList(values())
        .parallelStream()
        .filter(VerifierType::isAutoExecutable)
        .filter(v -> !v.isMultiTable() || includeMultiTable)
        .forEach(
            verifierType ->
                verifierType
                    .getResult(dmnDecisionList, dmnRuleMap, interdeterminismRules)
                    .ifPresent(builder::addVerifierResult));
    Interdeterminism.getResult(dmnDecisionList, dmnRuleMap, interdeterminismRules)
        .ifPresent(builder::addVerifierResult);
    return builder.build();
  }
}
