package de.unikoblenz.fgbks.dmn.core.verifier;

import de.unikoblenz.fgbks.dmn.core.dmnhelper.DmnRuleMap;
import de.unikoblenz.fgbks.dmn.core.dmnhelper.Type;
import de.unikoblenz.fgbks.dmn.core.dmnhelper.Value;
import de.unikoblenz.fgbks.dmn.core.models.RuleIdentifier;
import de.unikoblenz.fgbks.dmn.core.models.VerificationResult;
import de.unikoblenz.fgbks.dmn.core.models.VerifierResult;
import de.unikoblenz.fgbks.dmn.core.models.VerifierType;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Queue;
import org.camunda.bpm.dmn.engine.DmnDecision;

/** An abstract Verifier for dmn decisions. */
public abstract class AbstractVerifier {

  private VerifierType verifierType;
  private VerifierResult.Builder verifierResultBuilder;
  protected List<DmnDecision> dmnDecisionList;
  protected DmnRuleMap dmnRuleMap;
  protected Queue<VerificationResult> interdeterminismRules;

  /** It's mandatory to set the VerifierType before calling findValidationErrors. */
  public AbstractVerifier withVerifierType(VerifierType verifierType) {
    this.verifierType = verifierType;
    interdeterminismRules = new LinkedList<>();
    return this;
  }

  public Queue<VerificationResult> getInterdeterminismRules() {
    return interdeterminismRules;
  }

  public void setInterdeterminismRules(Queue<VerificationResult> interdeterminismRules) {
    this.interdeterminismRules = interdeterminismRules;
  }

  /**
   * Start the calculation of the verifier
   *
   * @param dmnDecisionList the list DmnDecision to calculate the verifications
   * @param dmnRuleMap a dmnRuleMap, generated from the dmnDecisionList
   * @return the VerifierResult
   */
  public final VerifierResult findValidationErrors(
      final List<DmnDecision> dmnDecisionList, final DmnRuleMap dmnRuleMap) {
    Objects.requireNonNull(verifierType);
    this.dmnDecisionList = new ArrayList<>(Objects.requireNonNull(dmnDecisionList));
    this.dmnRuleMap = Objects.requireNonNull(dmnRuleMap);
    verifierResultBuilder = VerifierResult.getBuilder().withVerifierType(verifierType);
    beforeVerifyDecision();
    for (DmnDecision d : dmnDecisionList) {
      verifyDecision(d);
    }
    afterVerifyDecision();
    return verifierResultBuilder.build();
  }

  /**
   * Start the calculation of the verifier
   *
   * @param dmnDecisionList the list DmnDecision to calculate the verifications
   * @return the VerifierResult
   */
  public final VerifierResult findValidationErrors(final List<DmnDecision> dmnDecisionList) {
    return findValidationErrors(
        dmnDecisionList, DmnRuleMap.createMapFromDmn(Objects.requireNonNull(dmnDecisionList)));
  }

  protected abstract void beforeVerifyDecision();

  protected abstract void verifyDecision(DmnDecision dmnDecision);

  protected abstract void afterVerifyDecision();

  protected final void addVerification(VerificationResult verificationResult) {
    verifierResultBuilder.addVerificationResult(Objects.requireNonNull(verificationResult));
  }

  protected boolean checkDifferentConclusion(
      String decisionKey, List<RuleIdentifier> currentRuleIdentifiers) {
    List<Type> outputCols =
        dmnRuleMap.getAllOutputTypesFromDecisionKey(Objects.requireNonNull(decisionKey));
    for (Type outputType : outputCols) {
      List<Value> values =
          dmnRuleMap.getValuesFromOutputType(
              outputType, Objects.requireNonNull(currentRuleIdentifiers));
      if (values.stream().map(Value::getBoundary).distinct().limit(2).count() > 1) {
        return true;
      }
    }
    return false;
  }
}
