package de.unikoblenz.fgbks.dmn.core.verifier;

import de.unikoblenz.fgbks.dmn.core.dmnhelper.Boundary;
import de.unikoblenz.fgbks.dmn.core.dmnhelper.Type;
import de.unikoblenz.fgbks.dmn.core.dmnhelper.Value;
import de.unikoblenz.fgbks.dmn.core.models.RuleIdentifier;
import de.unikoblenz.fgbks.dmn.core.models.VerificationResult;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import org.camunda.bpm.dmn.engine.DmnDecision;

public class MissingVerifier extends AbstractVerifier {

  @Override
  protected void beforeVerifyDecision() {
    // Nothing
  }

  @Override
  protected void verifyDecision(DmnDecision d) {
    if (d.isDecisionTable()) {
      List<Type> inputs = new ArrayList<>(dmnRuleMap.getAllInputTypesFromDecisionKey(d.getKey()));
      List<Value[]> missingRules = checkForMissingRules(inputs, dmnRuleMap.getRulesFromDMNTable(d));
      // add errors for missing intervals
      for (Value[] missingInt : missingRules) {
        StringBuilder sb = new StringBuilder("In table ");
        sb.append(d.getName()); // TODO: add real name
        sb.append(" the following rule is not defined: (");
        sb.append(
            Arrays.stream(missingInt)
                .map(Value::getOrgExpression)
                .collect(Collectors.joining("), (")));
        sb.append(")");
        addVerification(
            VerificationResult.getBuilder()
                .addRules(
                    Arrays.stream(missingInt)
                        .map(Value::getRuleIdentifier)
                        .collect(Collectors.toList()))
                .withMessage(sb.toString())
                .build());
      }
    }
  }

  private List<Value[]> checkForMissingRules(
      List<Type> inputs, List<RuleIdentifier> rulesFromDMNTable) {
    // check, if missingIntervals are empty
    List<Value[]> missingRules = new ArrayList<>();
    missingRules.add(constructMissingRule(inputs));
    // Check new Missing Rules for each rule
    for (RuleIdentifier ruleIdentifier : rulesFromDMNTable) {
      checkForMissingRules(
          inputs, dmnRuleMap.getRuleInputValuesFromIdentifier(ruleIdentifier), missingRules);
    }
    mergeRules(missingRules);
    return missingRules;
  }

  private void checkForMissingRules(
      List<Type> inputs, List<Value> values, List<Value[]> missingRules) {
    // search the missing rules, which are in contact with the current rule
    List<Value[]> inContactRules = new ArrayList<>();
    for (Value[] missingRule : missingRules) {
      if (isInContact(missingRule, values)) {
        inContactRules.add(missingRule);
      }
    }
    // remove found rules from current missingRules
    missingRules.removeAll(inContactRules);
    // Build new missing rules from in contact rules
    List<Value[]> newMissingRules = new ArrayList<>();
    for (Value[] missingRuleValues : inContactRules) {
      newMissingRules.addAll(constructNewMissingRules(inputs, missingRuleValues, values));
    }
    // mergeRules(newMissingRules);
    missingRules.addAll(newMissingRules);
    // mergeRules(missingRules);
  }

  private boolean isInContact(Value[] missingRule, List<Value> values) {
    for (int i = 0; i < missingRule.length; i++) {
      if (missingRule[i].getBoundary().isNotInContact(values.get(i).getBoundary())) {
        return false;
      }
    }
    return true;
  }

  private List<Value[]> constructNewMissingRules(
      List<Type> inputs, Value[] missingRuleValues, List<Value> values) {
    List<Value[]> newMissingRules = new ArrayList<>();
    for (int i = 0; i < missingRuleValues.length; i++) {
      // change bound for i-1
      if (i != 0) {
        Value mRV = missingRuleValues[i - 1];
        if (mRV.getType().getDataType().isNumeric()) {
          Boundary mRB = mRV.getBoundary();
          Boundary vB = values.get(i - 1).getBoundary();
          Optional<Boundary> b = vB.createBoundaryIntersection(mRB);
          if (b.isPresent()) {
            missingRuleValues[i - 1] =
                new Value(
                    b.get(),
                    b.get().toString(),
                    RuleIdentifier.getBuilder()
                        .withRowNumber(999)
                        .withDecision(inputs.get(i).getDecision())
                        .withRuleId("newMissingRule")
                        .build(),
                    inputs.get(i - 1));
          }
        } else {
          // TODO: STRING
        }
      }

      // calc next missing intervals: left + right
      Value currentValue = values.get(i);
      if (currentValue.getType().getDataType().isNumeric()) {
        // left
        addMissingRules(
            i,
            inputs,
            missingRuleValues[i]
                .getBoundary()
                .createBoundaryWithLowerBounds(currentValue.getBoundary()),
            missingRuleValues,
            newMissingRules);

        // right
        addMissingRules(
            i,
            inputs,
            missingRuleValues[i]
                .getBoundary()
                .createBoundaryWithUpperBounds(currentValue.getBoundary()),
            missingRuleValues,
            newMissingRules);

      } else {
        // TODO: String
      }
    }
    return newMissingRules;
  }

  private void addMissingRules(
      int i,
      List<Type> inputs,
      Optional<Boundary> b,
      Value[] missingRuleValues,
      List<Value[]> newMissingRules) {
    if (b.isPresent()) {
      Value[] copy = cloneValueArray(missingRuleValues);
      copy[i] =
          new Value(
              b.get(),
              b.get().toString(),
              RuleIdentifier.getBuilder()
                  .withRowNumber(999)
                  .withDecision(inputs.get(i).getDecision())
                  .withRuleId("newMissingRule")
                  .build(),
              inputs.get(i));
      newMissingRules.add(copy);
    }
  }

  private Value[] cloneValueArray(Value[] org) {
    Value[] newAry = new Value[org.length];
    for (int i = 0; i < org.length; i++) {
      Boundary b = null;
      try {
        b = (Boundary) org[i].getBoundary().clone();
      } catch (CloneNotSupportedException e) {
        // nothing
      }
      newAry[i] =
          new Value(
              b,
              b.toString(),
              RuleIdentifier.getBuilder()
                  .withRowNumber(999)
                  .withDecision(org[i].getType().getDecision())
                  .withRuleId("newMissingRule")
                  .build(),
              org[i].getType());
    }
    return newAry;
  }

  private Value[] constructMissingRule(List<Type> inputs) {
    Value[] newMissing = new Value[inputs.size()];
    for (int i = 0; i < newMissing.length; i++) {
      newMissing[i] =
          new Value(
              Boundary.parseBoundaryFromNullValue(),
              "-",
              RuleIdentifier.getBuilder()
                  .withRowNumber(999)
                  .withDecision(inputs.get(i).getDecision())
                  .withRuleId("newMissingRule")
                  .build(),
              inputs.get(i));
    }
    return newMissing;
  }

  private void mergeRules(List<Value[]> missingRules) {
    boolean found = true;
    while (found) {
      List<Value[]> removeRules = new ArrayList<>();
      found = false;
      for (int i = 0; i < missingRules.size(); i++) {
        for (int u = i + 1; u < missingRules.size(); u++) {
          int foundCombination = -1;
          Optional<Boundary> newOBoundary;
          Boundary newBoundary = null;

          for (int x = 0; x < missingRules.get(i).length; x++) {
            Value v1 = missingRules.get(i)[x];
            Value v2 = missingRules.get(u)[x];
            newOBoundary = v1.getBoundary().combineWith(v2.getBoundary());
            if (v1.getBoundary().isEqualsTo(v2.getBoundary())) {

            } else if (newOBoundary.isPresent() && foundCombination == -1) {
              newBoundary = newOBoundary.get();
              foundCombination = x;
            } else {
              foundCombination = -1;
              break;
            }
          }
          if (foundCombination >= 0) {
            found = true;
            missingRules.get(i)[foundCombination].setBoundary(newBoundary);
            removeRules.add(missingRules.get(u));
          }
        }
      }
      missingRules.removeAll(removeRules);
    }
  }

  @Override
  protected void afterVerifyDecision() {
    // Nothing
  }
}
