package de.unikoblenz.fgbks.dmn.core.verifier;

import de.unikoblenz.fgbks.dmn.core.dmnhelper.Type;
import de.unikoblenz.fgbks.dmn.core.dmnhelper.Value;
import de.unikoblenz.fgbks.dmn.core.models.RuleIdentifier;
import de.unikoblenz.fgbks.dmn.core.models.VerificationResult;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;
import org.camunda.bpm.dmn.engine.DmnDecision;

public class OverlappingVerifier extends AbstractVerifier {

  @Override
  protected void beforeVerifyDecision() {}

  @Override
  protected void verifyDecision(DmnDecision d) {
    if (d.isDecisionTable()) {
      List<Type> inputs = new ArrayList<>(dmnRuleMap.getAllInputTypesFromDecisionKey(d.getKey()));
      checkForOverlappingRules(inputs, 0, null, false, Collections.emptyList());
    }
  }

  protected void checkForOverlappingRules(
      List<Type> inputs,
      int i,
      List<RuleIdentifier> currentRuleIdentifiers,
      boolean hasOverlap,
      List<RuleIdentifier> subsubmtionRules) {
    if (i == inputs.size()) {
      // do nothing, if there was no real overlap found prev..
      if (hasOverlap) {
        boolean differentConclusions =
            checkDifferentConclusion(inputs.get(0).getDecisionKey(), currentRuleIdentifiers);
        VerificationResult.Builder vBuilder =
            VerificationResult.getBuilder()
                .withMessage(getMessageText(currentRuleIdentifiers, differentConclusions));
        vBuilder.addRules(currentRuleIdentifiers);
        VerificationResult vResult = vBuilder.build();
        if (differentConclusions) {
          interdeterminismRules.add(vResult);
        }
        addVerification(vResult);
      }
    } else {
      List<Value> currentBounds = new ArrayList<>();
      List<Value> sortedBounds =
          dmnRuleMap.getValuesFromInputType(inputs.get(i), currentRuleIdentifiers);
      sortedBounds.sort(Comparator.comparing(Value::getBoundary));
      int z = 0;
      boolean foundPOverlap = false;
      for (Value currentBound : sortedBounds) {
        z++;
        boolean foundOverlap = hasOverlap;
        // copy of current bounds
        List<Value> currentBoundsCpy = new ArrayList<>(currentBounds);
        // add current bound to overlapping bounds
        currentBounds.add(currentBound);
        // check, if current bound is overlapping an other bound in the current list of overlapping
        // rules
        boolean foundNotInContact = false;
        for (int x = 0; x < currentBounds.size(); x++) {
          Value bound = currentBounds.get(x);
          if (currentBound.getBoundary().isNotInContact(bound.getBoundary())) {
            currentBounds.remove(x--);
            foundNotInContact = true;
          } else {
            foundOverlap =
                foundOverlap || currentBound.getBoundary().isOverlapping(bound.getBoundary());
          }
        }
        if (foundNotInContact && currentBoundsCpy.size() > 1) {
          List<RuleIdentifier> newSubsumtionRules = searchSubsubmtionElement(currentBoundsCpy);
          checkForOverlappingRules(
              inputs,
              i + 1,
              currentBoundsCpy.stream().map(Value::getRuleIdentifier).collect(Collectors.toList()),
              hasOverlap
                  || foundPOverlap
                  || (newSubsumtionRules.size() > 0
                      && subsubmtionRules.size() > 0
                      && containsNoDuplicateRows(newSubsumtionRules, subsubmtionRules)),
              newSubsumtionRules);
          foundPOverlap = false;
        }
        if (z == sortedBounds.size() && currentBounds.size() > 1) {
          List<RuleIdentifier> newSubsumtionRules = searchSubsubmtionElement(currentBounds);
          // check if
          checkForOverlappingRules(
              inputs,
              i + 1,
              currentBounds.stream().map(Value::getRuleIdentifier).collect(Collectors.toList()),
              hasOverlap
                  || foundOverlap
                  || (newSubsumtionRules.size() > 0
                      && subsubmtionRules.size() > 0
                      && containsNoDuplicateRows(newSubsumtionRules, subsubmtionRules)),
              newSubsumtionRules);
        }
        foundPOverlap |= foundOverlap;
      }
    }
  }

  private List<RuleIdentifier> searchSubsubmtionElement(List<Value> values) {
    List<RuleIdentifier> subsubmtionRules = new ArrayList<>();
    outer:
    for (Value value : values) {
      boolean containsSubsumption = false;
      for (Value innerValue : values) {
        if (value.getBoundary().subsumes(innerValue.getBoundary())) {
          containsSubsumption = true;
        } else if (value.getBoundary().isNotInContact(innerValue.getBoundary())
            || value.getBoundary().isOverlapping(innerValue.getBoundary())) {
          continue outer;
        }
      }
      if (containsSubsumption) {
        subsubmtionRules.add(value.getRuleIdentifier());
      }
    }
    return subsubmtionRules;
  }

  private boolean containsNoDuplicateRows(List<RuleIdentifier> ri1, List<RuleIdentifier> ri2) {
    for (RuleIdentifier r1 : ri1) {
      for (RuleIdentifier r2 : ri2) {
        if (r1.getDecisionKey().equals(r2.getDecisionKey())
            && r1.getRowNumber().equals(r2.getRowNumber())) {
          return false;
        }
      }
    }
    return true;
  }

  @Override
  protected void afterVerifyDecision() {
    // Nothing
  }

  protected String getMessageText(
      List<RuleIdentifier> currentRuleIdentifiers, boolean isDifferentConclusion) {
    StringBuilder sb = new StringBuilder();
    sb.append("In table ");
    sb.append(currentRuleIdentifiers.get(0).getTableName());
    sb.append(" rules ");
    sb.append(
        currentRuleIdentifiers.stream()
            .map(c -> c.getRowNumber().toString())
            .collect(Collectors.joining(", ")));
    sb.append(" are overlapping.");
    if (isDifferentConclusion) {
      sb.append(" The output has different conclusions!");
    } else {
      sb.append(" The output is the same.");
    }
    return sb.toString();
  }
}
