package de.unikoblenz.fgbks.dmn.core.verifier;

import de.unikoblenz.fgbks.dmn.core.dmnhelper.Type;
import de.unikoblenz.fgbks.dmn.core.dmnhelper.Value;
import de.unikoblenz.fgbks.dmn.core.models.RuleIdentifier;
import de.unikoblenz.fgbks.dmn.core.models.VerificationResult;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import org.camunda.bpm.dmn.engine.DmnDecision;

public class PartialReductionVerifier extends AbstractVerifier {

  @Override
  protected void beforeVerifyDecision() {
    // Nothing
  }

  @Override
  protected void verifyDecision(DmnDecision d) {
    if (d.isDecisionTable()) {
      List<List<RuleIdentifier>> identicalOutputCluster = findIdenticalOutputs(d.getKey());
      List<Type> inputs = new ArrayList<>(dmnRuleMap.getAllInputTypesFromDecisionKey(d.getKey()));
      identicalOutputCluster.stream() // do: parallelStream() ?
          .forEach(c -> findPartialReduction(inputs, 0, c, false));
    }
  }

  protected List<List<RuleIdentifier>> findIdenticalOutputs(String dmnKey) {
    List<List<RuleIdentifier>> clusters = new ArrayList<>();
    List<Type> outputs = new ArrayList<>(dmnRuleMap.getAllOutputTypesFromDecisionKey(dmnKey));

    List<RuleIdentifier> rowsIds =
        dmnRuleMap.getValuesFromType(outputs.get(0)).stream()
            .map(Value::getRuleIdentifier)
            .collect(Collectors.toList());
    int rowsCount = dmnRuleMap.getValuesFromType(outputs.get(0)).size();
    for (int i = 0; i < rowsCount; i++) {
      List<RuleIdentifier> cluster = new ArrayList<>();
      cluster.add(rowsIds.get(i));
      inner:
      for (int u = i + 1; u < rowsCount; u++) {
        for (List<RuleIdentifier> clu : clusters) {
          if (clu.contains(rowsIds.get(u))) {
            continue inner;
          }
        }
        if (checkRulesHasIdenticalOutput(dmnKey, rowsIds.get(i), rowsIds.get(u))) {
          cluster.add(rowsIds.get(u));
        }
      }
      if (cluster.size() > 1) {
        clusters.add(cluster);
      }
    }
    return clusters;
  }

  private boolean checkRulesHasIdenticalOutput(
      String dmnKey, RuleIdentifier r1, RuleIdentifier r2) {
    List<Type> outputs = new ArrayList<>(dmnRuleMap.getAllOutputTypesFromDecisionKey(dmnKey));
    for (Type type : outputs) {
      Value v1;
      Value v2;
      try {
        v1 = dmnRuleMap.getValuesFromOutputType(type, Collections.singletonList(r1)).get(0);
        v2 = dmnRuleMap.getValuesFromOutputType(type, Collections.singletonList(r2)).get(0);
      } catch (Exception e) {
        return false;
      }
      if (!v1.getBoundary().isEqualsTo(v2.getBoundary())) {
        return false;
      }
    }
    return true;
  }

  protected void findPartialReduction(
      List<Type> inputs,
      int i,
      List<RuleIdentifier> identicalOutputCluster,
      boolean hasCombination) {
    if (i == inputs.size()) {
      if (hasCombination) {
        VerificationResult.Builder vBuilder =
            VerificationResult.getBuilder().withMessage(getMessageText(identicalOutputCluster));
        vBuilder.addRules(identicalOutputCluster);
        addVerification(vBuilder.build());
      }
    } else {
      List<Value> rules = dmnRuleMap.getValuesFromInputType(inputs.get(i), identicalOutputCluster);
      List<List<RuleIdentifier>> nClusters = new ArrayList<>();
      List<Boolean> combinationCluster = new ArrayList<>();
      for (int i1 = 0; i1 < rules.size(); i1++) {
        Value currentRule = rules.get(i1);
        // 1. Find combinations
        if (!hasCombination) {
          for (int i2 = i1 + 1; i2 < rules.size(); i2++) {
            Value currentRuleInner = rules.get(i2);
            if (currentRule != currentRuleInner) {
              if (currentRule
                  .getBoundary()
                  .combineWith(currentRuleInner.getBoundary())
                  .isPresent()) {
                List<RuleIdentifier> cluster = new ArrayList<>();
                cluster.add(currentRule.getRuleIdentifier());
                cluster.add(currentRuleInner.getRuleIdentifier());
                nClusters.add(cluster);
                combinationCluster.add(true);
              }
            }
          }
        }
        // 2. Find equal values
        for (int i2 = i1 + 1; i2 < rules.size(); i2++) {
          Value currentRuleInner = rules.get(i2);
          if (currentRule != currentRuleInner) {
            if (currentRule.getBoundary().isEqualsTo(currentRuleInner.getBoundary())) {
              List<RuleIdentifier> cluster = new ArrayList<>();
              cluster.add(currentRule.getRuleIdentifier());
              cluster.add(currentRuleInner.getRuleIdentifier());
              nClusters.add(cluster);
              combinationCluster.add(false);
            }
          }
        }
      }
      // nex col
      for (int x = 0; x < combinationCluster.size(); x++) {
        if (nClusters.get(x).size() > 1) {
          findPartialReduction(
              inputs, i + 1, nClusters.get(x), combinationCluster.get(x) || hasCombination);
        }
      }
    }
  }

  @Override
  protected void afterVerifyDecision() {
    // Nothing
  }

  protected String getMessageText(List<RuleIdentifier> ruleIdentifiers) {
    StringBuilder sb = new StringBuilder("Rules ");
    sb.append(
        ruleIdentifiers.stream()
            .map(c -> c.getRowNumber().toString())
            .collect(Collectors.joining(" and ")));
    sb.append(" in table ");
    sb.append(ruleIdentifiers.get(0).getTableName());
    sb.append(" can be combined.");
    return sb.toString();
  }
}
