package de.unikoblenz.fgbks.dmn.core.verifier;

import de.unikoblenz.fgbks.dmn.core.dmnhelper.Type;
import de.unikoblenz.fgbks.dmn.core.dmnhelper.Value;
import de.unikoblenz.fgbks.dmn.core.models.RuleIdentifier;
import de.unikoblenz.fgbks.dmn.core.models.VerificationResult;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import org.camunda.bpm.dmn.engine.DmnDecision;

public class SubsumptionVerifier extends AbstractVerifier {

  @Override
  protected void beforeVerifyDecision() {}

  @Override
  protected void verifyDecision(DmnDecision d) {
    if (d.isDecisionTable()) {
      List<Type> inputs = new ArrayList<>(dmnRuleMap.getAllInputTypesFromDecisionKey(d.getKey()));
      checkForSubsumptionRules(inputs, 0, null, false, null);
    }
  }

  protected void checkForSubsumptionRules(
      List<Type> inputs,
      int i,
      List<RuleIdentifier> currentRuleIdentifiers,
      boolean hasSubsumption,
      List<Value> currentRootSubsumptionElements) {
    if (i == inputs.size()) {
      // finish, if subsumption was found previously, than add the rules
      if (hasSubsumption && currentRuleIdentifiers.size() > currentRootSubsumptionElements.size()) {
        boolean differentConclusions =
            checkDifferentConclusion(inputs.get(0).getDecisionKey(), currentRuleIdentifiers);
        VerificationResult.Builder vBuilder =
            VerificationResult.getBuilder()
                .withMessage(
                    getMessageText(
                        currentRuleIdentifiers,
                        currentRootSubsumptionElements,
                        differentConclusions));
        vBuilder.addRules(currentRuleIdentifiers);
        VerificationResult vResult = vBuilder.build();
        if (differentConclusions) {
          interdeterminismRules.add(vResult);
        }
        addVerification(vResult);
      }
    } else {
      // get all input values from the current column, and filter with the prev. iteration
      List<Value> selectedBounds =
          new ArrayList<>(dmnRuleMap.getValuesFromInputType(inputs.get(i), currentRuleIdentifiers));

      List<Set<Value>> clusters = new ArrayList<>(); // clusters of subsumptions
      List<Boolean> subsumptions =
          new ArrayList<>(); // is the value in selectedBounds a subsubmption?
      List<List<Value>> subsumptionValue =
          new ArrayList<>(); // save the subsumption value of the current cluster
      // if previously a rule was found, which has a subsumption, than it must be also the element,
      // which subsume the other elements.
      // First, build clusters, which contain a subsumption set
      if (currentRootSubsumptionElements != null) {
        // get the value from the subsumption rule
        for (Value sElement : currentRootSubsumptionElements) {
          List<Value> subsumptionVals = new ArrayList<>();
          Value nValue =
              dmnRuleMap
                  .getValuesFromInputType(
                      inputs.get(i), Collections.singletonList(sElement.getRuleIdentifier()))
                  .get(0);
          Set<Value> c1 = new HashSet<>();
          boolean foundSubsumption = false;
          for (Value cb2 : selectedBounds) {
            boolean fs = nValue.getBoundary().subsumes(cb2.getBoundary());
            foundSubsumption |= fs;
            if (fs) {
              c1.add(cb2);
            } else if (nValue.getBoundary().isEqualsTo(cb2.getBoundary())) {
              c1.add(cb2);
              if (currentRootSubsumptionElements.contains(cb2)) {
                subsumptionVals.add(cb2);
              }
            }
          }
          if (c1.size() > 1) {
            clusters.add(c1);
            subsumptions.add(foundSubsumption);
            subsumptionValue.add(subsumptionVals);
          }
        }
      } else {
        // no subsumption was found yet
        // --> build all clusters, where one rule subsume other rules
        for (Value cb1 : selectedBounds) {
          Set<Value> c1 = new HashSet<>();
          boolean foundSubsumption = false;
          for (Value cb2 : selectedBounds) {
            boolean fs = cb1.getBoundary().subsumes(cb2.getBoundary());
            foundSubsumption |= fs;
            if (fs) {
              c1.add(cb2);
            }
          }
          List<Value> subsumptionVals = new ArrayList<>();
          for (Value cb2 : selectedBounds) {
            if (cb1.getBoundary().isEqualsTo(cb2.getBoundary())) {
              c1.add(cb2);
              if (foundSubsumption) {
                subsumptionVals.add(cb2);
              }
            }
          }
          if (c1.size() > 1) {
            clusters.add(c1);
            subsumptions.add(foundSubsumption);
            if (foundSubsumption) {
              subsumptionValue.add(subsumptionVals);
            } else {
              subsumptionValue.add(null);
            }
          }
        }
      }
      // check for subsets, that includes other subsets and remove them
      a:
      for (int x = 0; x < clusters.size(); x++) {
        if (clusters.get(x) != null) {
          Set<Value> c1 = clusters.get(x);
          for (int y = 0; y < clusters.size(); y++) {
            if (clusters.get(y) != null) {
              Set<Value> c2 = clusters.get(y);
              if (x != y
                  && ((subsumptionValue.get(x) != null && subsumptionValue.get(y) != null)
                      || ((subsumptionValue.get(x) == null && subsumptionValue.get(y) == null)
                          && c1.containsAll(c2)
                          && c2.containsAll(c1)))) {
                if (c1.containsAll(c2)) {
                  clusters.set(y, null);
                }
              }
            }
          }
        }
      }

      // recursively search next row for subsumption with all clusters
      for (int x = 0; x < clusters.size(); x++) {
        if (clusters.get(x) != null) {
          Set<Value> vals = clusters.get(x);
          checkForSubsumptionRules(
              inputs,
              i + 1,
              vals.stream().map(Value::getRuleIdentifier).collect(Collectors.toList()),
              subsumptions.get(x) || hasSubsumption,
              subsumptionValue.get(x));
        }
      }
    }
  }

  @Override
  protected void afterVerifyDecision() {}

  protected String getMessageText(
      List<RuleIdentifier> currentRuleIdentifiers,
      List<Value> subsumptionRules,
      boolean isDifferentConclusion) {
    StringBuilder sb = new StringBuilder();
    sb.append("In table ");
    sb.append(subsumptionRules.get(0).getRuleIdentifier().getTableName());
    sb.append(" rule");
    if (subsumptionRules.size() > 1) {
      sb.append("s");
    }
    sb.append(" ");
    sb.append(
        subsumptionRules.stream()
            .map(c -> c.getRuleIdentifier().getRowNumber().toString())
            .sorted()
            .collect(Collectors.joining(", ")));
    sb.append(" subsumes rule");
    if (currentRuleIdentifiers.size() > 2) {
      sb.append("s");
    }
    sb.append(" ");
    sb.append(
        currentRuleIdentifiers.stream()
            .filter(
                r ->
                    !subsumptionRules.stream()
                        .map(Value::getRuleIdentifier)
                        .collect(Collectors.toList())
                        .contains(r))
            .map(c -> c.getRowNumber().toString())
            .sorted()
            .collect(Collectors.joining(", ")));
    sb.append(".");
    if (isDifferentConclusion) {
      sb.append(" The output has different conclusions!");
    } else {
      sb.append(" The output is the same.");
    }
    return sb.toString();
  }
}
